#ifndef __SD_CHART__
#define __SD_CHART__

#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm> // sort, max_element
#include <queue>

#include "util.h"
#include "cell.h"
#include "kbest.h"
#include "double_array.h"

namespace SD {

class Chart {
  private:
    static const Real EPSILON;
    typedef std::priority_queue< Real,std::vector<Real>,std::greater<Real> > Agenda;
    unsigned int  L_;
    unsigned int  D_;
    Cell* cell_;
    std::vector<Cell*> first_;
    std::vector<Cell*> last_;
    Agenda A_;
    KBest<Cell*,Real> kbest_;
  public:
    Chart();
    Chart( const Chart& obj );
    ~Chart();
    Chart& operator=( const Chart& obj );
    void set_size( unsigned int L, unsigned int D );
    void viterbi_init( unsigned int n );
    void staggered_init( unsigned int n );
    template<class Model> void carpediem_init( unsigned int n, const Model* model );
    Cell* first( unsigned int n );
    const Cell* first( unsigned int n ) const;
    Cell* last( unsigned int n );
    const Cell* last( unsigned int n ) const;
    template<class Model> void expand( unsigned int i, const Model* model );
    void prune( unsigned int i, Real lb );
    void prune2( unsigned int i, unsigned int k );
    Real traverse_backlink1( Label* y, unsigned int n ) const;
    template<class Model> Real traverse_backlink2( Label* y, unsigned int n, const Model* model ) const;
    void print( FILE* is, unsigned int n, const DoubleArray& indexer ) const;
};
const Real Chart::EPSILON = 10e-3;


Chart::Chart() {

    cell_  = NULL;
}


Chart::~Chart() {

    delete[] cell_;
}


inline void Chart::set_size( unsigned int L, unsigned int D ) {

    delete[] cell_;
    cell_  = new Cell[MAX_LENGTH*L];
    first_.resize( MAX_LENGTH );
    last_.resize( MAX_LENGTH );
    L_ = L;
    D_ = D;
}


inline void Chart::viterbi_init( unsigned int n ) {
    
    Cell* c = cell_;
    for ( unsigned int i = 0; i < n; ++i ) {
	for ( Label y = 0; y < L_; ++y ) {
	    c->label = y;
	    if ( y == 0 ) {
		c->prev = NULL;
		first_[i] = c;
	    }else {
		c->prev = c-1;
	    }
	    if ( y == L_-1 ) {
		c->next = NULL;
		last_[i] = c;
	    }else {
		c->next = c+1;
	    }
	    ++c;
	}
    }
}


inline void Chart::staggered_init( unsigned int n ) {

    Cell* c = cell_;
    for ( unsigned int i = 0; i < n; ++i ) {
	// One active label
	first_[i] = c;
	c->label  = 0;
	c->next   = c + 1;
	c->prev   = NULL;
	++c;
	
	// One degenerate label
	last_[i] = c;
	c->label = L_;
	c->next  = NULL;
	c->prev  = c - 1;
	c += (L_-1);
    }
}


template<class Model>
inline void Chart::carpediem_init( unsigned int n, const Model* model ) {

    Cell* c = cell_;
    for ( unsigned int i = 0; i < n; ++i ) {
	for ( Label label = 0; label < L_; ++label ) {
	    c->label = label;
	    c->alpha = model->node_score( i, label );
	    ++c;
	}
	c -= L_;
	std::sort( c, c + L_, cmp_cell() );
	
	for ( unsigned int j = 0; j < L_; ++j ) {
	    c->alpha = MIN_VALUE;   // use to check open/close
	    if ( j == L_-1 ) {
		c->next  = NULL;
		last_[i] = c;
	    }else {
		c->next = c+1;
	    }
	    if ( j == 0 ) {
		c->prev   = NULL;
		first_[i] = c;
	    }else {
		c->prev = c-1;
	    }
	    ++c;
	}
    }
}


inline void Chart::prune( unsigned int i, Real lb ) {

    for ( Cell* c = first_[i]; c != NULL; c = c->next ) {
	if ( c->alpha + c->beta + EPSILON < lb ) {
	    Cell* c_prev = c->prev;
	    Cell* c_next = c->next;
	    if ( c_prev == NULL ) {
		if ( c_next == NULL ) {
		    fprintf( stderr, HERE "An error occurred in pruning\n" );
		    exit( FAILURE );
		}else {
		    c_next->prev = NULL;
		    first_[i] = c_next;
		}
	    }else if ( c_next == NULL ) {
		c_prev->next = NULL;
		last_[i] = c_prev;
	    }else {
		c_prev->next = c_next;
		c_next->prev = c_prev;
	    }
	}
    }
}


inline void Chart::prune2( unsigned int i, unsigned int k ) {

    /*
    kbest_.init( k );
    for ( Cell* c = first_[i]; c != NULL; c = c->next ) {
	kbest_.push( c, c->alpha );
    }
    Real lb = (kbest_.end()-1)->second;
    */
    
    /*
    for ( KBest<Cell*,Real>::iterator it = kbest_.begin(); it != kbest_.end(); ++it ) {
	std::cerr << "\t" << it->second;
    }
    std::cerr << std::endl;
    */

    while ( A_.empty() == false ) {
	A_.pop();
    }
    for ( Cell* c = first_[i]; c != NULL; c = c->next ) {
	if ( A_.size() == k ) {
	    if ( A_.top() < c->alpha ) {
		A_.pop();
		A_.push( c->alpha );
	    }
	}else {
	    A_.push( c->alpha );
	}
    }
    Real lb = A_.top();

    //std::cerr << lb << "\t" << lb2 << std::endl;
    
    for ( Cell* c = first_[i]; c != NULL; c = c->next ) {
	if ( c->alpha < lb ) {
	    Cell* c_prev = c->prev;
	    Cell* c_next = c->next;
	    if ( c_prev == NULL ) {
		c_next->prev = NULL;
		first_[i] = c_next;
	    }else if ( c_next == NULL ) {
		c_prev->next = NULL;
		last_[i] = c_prev;
	    }else {
		c_prev->next = c_next;
		c_next->prev = c_prev;
	    }
	}
    }
}


template<class Model> 
inline void Chart::expand( unsigned int i, const Model* model ) {

    // Memorize the last cell (= degenerate label)
    Cell* last  = last_[i];
    Label label = last->label;
    Real  alpha = last->alpha;
    Real  beta  = last->beta;
    Cell* prev  = last->prev;
	
    // Expand active labels
    if ( label+1 < L_+D_ ) {
	Cell* c = last;
	Label y_first = pow2( label-L_ );
	Label y_last = 2*y_first;
	for ( Label y = y_first; y < y_last; ++y ) {
	    c->label = y;
	    c->alpha = alpha + model->node_score( i, y ) - model->node_score( i, label );
	    c->beta  = beta;
	    if ( y == y_first ) {
		c->prev = prev;
	    }else {
		c->prev = c - 1;
	    }
	    c->next = c + 1;
	    ++c;
	}
	
	// degenerate label
	c->label = label + 1;
	c->alpha = alpha + model->node_score( i, c->label ) - model->node_score( i, label );
	c->beta  = beta;
	c->next  = NULL;
	c->prev  = c - 1;
	last_[i] = c;
    }else {
	Cell* c = last;
	Label y_first = pow2( label-L_ );
	for ( Label y = y_first; y < L_; ++y ) {
	    c->label = y;
	    c->alpha = alpha + model->node_score( i, y ) - model->node_score( i, label );
	    c->beta  = beta;
	    if ( y == y_first ) {
		c->prev = prev;
	    }else {
		c->prev = c - 1;
	    }
	    if ( y == L_ - 1 ) {
		c->next = NULL;
		last_[i] = c;
	    }else {
		c->next = c + 1;
		++c;
	    }
	}
    }
}


inline Real Chart::traverse_backlink1( Label* y, unsigned int n ) const {

    const Cell* best_c = NULL;
    Real best_score    = MIN_VALUE;

    for ( const Cell* c = first_[n-1]; c != NULL; c = c->next ) {
	Real score = c->alpha;
	if ( best_score < score ) {
	    best_c     = c;
	    best_score = score;
	}
    }

    y[n-1] = best_c->label;
    for ( unsigned int i = n-1; 1 <= i; --i ) {
	best_c = best_c->backlink;
	y[i-1] = best_c->label;
    }
    return best_score;
}


template<class Model> 
inline Real Chart::traverse_backlink2( Label* y, unsigned int n, const Model* model ) const {

    const Cell* best_c = NULL;
    Real best_score    = MIN_VALUE;
    
    for ( const Cell* c = first_[0]; c != NULL; c = c->next ) {
	Real score = model->node_score( 0, c->label ) + c->beta;
	if ( best_score < score ) {
	    best_c = c;
	    best_score = score;
	}
    }
    
    for ( unsigned int i = 0; i < n; ++i ) {
	y[i]   = best_c->label;
	best_c = best_c->backlink;
    }
    return best_score;
}



inline Cell* Chart::first( unsigned int i ) {
    
    return first_[i];
}


inline const Cell* Chart::first( unsigned int i ) const {
    
    return first_[i];
}


inline Cell* Chart::last( unsigned int i ) {
    
    return last_[i];
}


inline const Cell* Chart::last( unsigned int i ) const {
    
    return last_[i];
}


inline void Chart::print( FILE* is, unsigned int n, const DoubleArray& indexer ) const {

    unsigned int max_length = 8;
    char format[128];
    sprintf( format, "%s%ds", "%", max_length );
    
    for ( unsigned int i = 0; i < n; ++i ) {
	for ( Cell* c = first_[i]; c != NULL; c = c->next ) {
	    if ( c->label < L_ ) {
		fprintf( is, format, indexer.key(c->label) );
	    }else {
		fprintf( is, format, "*" );
	    }
	    fprintf( is, "(%5.3f)", c->beta );
	}
	fprintf( is, "\n" );
    }
}

}

#endif
