#ifndef __SD_DECODER__
#define __SD_DECODER__

#include <iostream>
#include <fstream>
#include <vector>
#include <queue>
#include <functional> // greater
#include <algorithm>  // sort

#include "util.h"
#include "cell.h"
#include "chart.h"
#include "timer.h"

namespace SD {

enum Algorithm {
    STAGGERED, VITERBI, BEAM, DET, CARPEDIEM
};

enum Expansion {
    FLAT, CEXP
};

enum Initialization {
    NEG_INF, HEURISTIC
};

struct DecodingOption{
    Algorithm      algorithm;
    unsigned int   beam_width;
    bool           pruning;
    Expansion      expansion;
    Initialization initialization;
    DecodingOption() {
	algorithm      = STAGGERED;
	beam_width     = 1;
	pruning        = true;
	expansion      = CEXP;
	initialization = HEURISTIC;
    }
};


template<class Model>
class Decoder {
  private:
    typedef typename Model::Input Input;
    TimerPool pool_;
    Timer* timer1_;
    Timer* timer2_;
    
    unsigned int L_;       // # of original labels
    unsigned int D_;       // # of degenerate labels
    Model* model_; // Model    
    DecodingOption option_;

    Chart chart2_;
    static const Real EPSILON;
    std::vector<Cell>  chart_;
    std::vector<Cell*> first_;
    std::vector<Cell*> last_;
    
    // For beam serach
    std::priority_queue< Real,std::vector<Real>,std::greater<Real> > A_;
    
    // For CarpeDiem
    std::vector< std::pair<Label,Real> > buff_;
    std::vector<Real> gamma_;
    
    // For staggered decoding
    unsigned int iter_num_;   // # iteration
    Real lb_;         // Lower bound value
    
    // Logging
    unsigned int  edge_num_;
    float edge_rate_;
  public:
    Decoder();
    Decoder( const Decoder& obj );
    ~Decoder();
    Decoder& operator=( const Decoder& obj );
    void init( unsigned int L, unsigned int D, Model* model, DecodingOption option );
    Real run( const Input* x, Label* y, unsigned int n );
    Real staggered( const Input* x, Label* y, unsigned int n );
    Real viterbi( const Input* x, Label* y, unsigned int n );
    Real carpediem( const Input* x, Label* y, unsigned int n );
    Real det( const Input* x, Label* y, unsigned int n );
    Real beam( const Input* x, Label* y, unsigned int n );
    
    // Debugging 
    void debug_chart( unsigned int i, std::vector<Label>& y ) const;
    unsigned int iter_num() const;
    float edge_rate() const;
  private:        
    void agenda_init();
    void agenda_add( Real score );
    void carpediem_open( const Input* x, unsigned int i, Cell* c );
    Real fw_staggered( const Input* x, Label* y, unsigned int n, bool prune_flag );
    Real bw_staggered( const Input* x, Label* y, unsigned int n, bool prune_flag );
    Real traverse_backlink1( Label* y, unsigned int n ) const;
    Real traverse_backlink2( Label* y, unsigned int n ) const;
    
    // Initialization
    void carpediem_init( unsigned int n );
    void viterbi_init( unsigned int n );
    void staggered_init( unsigned int n );
    
    // Change chart
    void expand_chart( unsigned int i );
    void prune_chart( unsigned int i );
    void prune_chart( unsigned int i, Real th );
};

template<class Model>
const Real Decoder<Model>::EPSILON = 10e-3;


template<class Model>
Decoder<Model>::Decoder() {

    timer1_ = pool_.push( "Preprosess@decoder.h" );
    timer2_ = pool_.push( "Decoding@decoder.h  " );
}

template<class Model>
Decoder<Model>::~Decoder() {

    pool_.print();
}


    
template<class Model>
void Decoder<Model>::init( unsigned int L, unsigned int D, Model* model, DecodingOption option ) {

    L_      = L;
    D_      = D;
    model_  = model;
    option_ = option;
    chart_.resize( MAX_LENGTH*L_ );
    first_.resize( MAX_LENGTH );
    last_.resize( MAX_LENGTH );
    gamma_.resize( MAX_LENGTH );

    chart2_.set_size( L, D );
}


template<class Model>
Real Decoder<Model>::run( const Input* x, Label* y, unsigned int n ) {

    if ( option_.algorithm == VITERBI ) {
	return viterbi( x, y, n );
    }else if ( option_.algorithm == BEAM ) {
	return beam( x, y, n );
    }else if ( option_.algorithm == DET ) {
	return det( x, y, n );
    }else if ( option_.algorithm == CARPEDIEM ) {
	return carpediem( x, y, n );
    }else {
	return staggered( x, y, n );
    }
}


/*
 *  Forward (left-to-right) Viterbi algorithm
 */
template<class Model>
inline Real Decoder<Model>::viterbi( const Input* x, Label* y, unsigned int n ) {

    if ( n == 0 ) return ZERO;

    timer1_->startTimer();
    iter_num_ = 0;
    model_->cache_node_score( x, n );
    chart2_.viterbi_init( n );
    timer1_->stopTimer();

    timer2_->startTimer();
    for ( Cell* c = chart2_.first(0); c != NULL; c = c->next ) {
	c->alpha    = model_->node_score( 0, c->label );
	c->backlink = NULL;
    }
    for ( unsigned int i = 1; i < n; ++i ) {
	Cell* c1 = chart2_.first(i-1);
	for ( Cell* c2 = chart2_.first(i); c2 != NULL; c2 = c2->next ) {
	    c2->alpha    = c1->alpha + model_->edge_score( c1->label, c2->label, x[i-1] ) + model_->node_score( i, c2->label );
	    c2->backlink = c1;
	}
	for ( c1 = c1->next; c1 != NULL; c1 = c1->next ) {
	    for ( Cell* c2 = chart2_.first(i); c2 != NULL; c2 = c2->next ) {
		Real score = c1->alpha + model_->edge_score( c1->label, c2->label, x[i-1] ) + model_->node_score( i, c2->label );
		if ( c2->alpha < score ) {
		    c2->alpha    = score;
		    c2->backlink = c1;
		}
	    }
	}
    }
    timer2_->stopTimer();
    
    return chart2_.traverse_backlink1( y, n );
}



/*
 *  CarpeDiem
 */
template<class Model>
inline Real Decoder<Model>::carpediem( const Input* x, Label* y, unsigned int n ) {

    if ( n == 0 ) return ZERO;

    model_->cache_node_score( x, n );
    carpediem_init( n );
    
#ifdef LOGGING
    edge_num_ = 0;
#endif
    for ( Cell* c = first_[0]; c != NULL; c = c->next ) {
	Real score = model_->node_score( 0, c->label );
	c->alpha = score;
	if ( gamma_[0] < score ) gamma_[0] = score;
#ifdef LOGGING	
	++edge_num_;
#endif
    }

    for ( unsigned int i = 1; i < n; ++i ) {
	//  Forward search strategy
	Cell* c = first_[i];
	carpediem_open( x, i, c );
	gamma_[i] = c->alpha;

	Real beta = gamma_[i-1] + model_->max_edge_score();
	for ( c = c->next; c != NULL; c = c->next ) {
	    if ( beta + model_->node_score( i, c->label ) < gamma_[i] ) break;
	    carpediem_open( x, i, c );
	    if ( gamma_[i] < c->alpha ) gamma_[i] = c->alpha;
	}
    }
#ifdef LOGGING	    
    edge_rate_ = static_cast<float>(edge_num_)/static_cast<float>(L_+(n-1)*L_*L_);
#endif
    
    return traverse_backlink1( y, n );
}


template<class Model>
inline void Decoder<Model>::carpediem_open( const Input* x, unsigned int i, Cell* c2 ) {

    Real beta = ZERO;
    if ( 2 <= i ) beta = gamma_[i-2] + model_->max_edge_score();
    
    for ( Cell* c1 = first_[i-1]; c1 != NULL; c1 = c1->next ) {
	if ( c1->alpha == MIN_VALUE ) {  // If c1 is closed
	    if ( c2->alpha < beta + model_->node_score( i-1, c1->label ) + model_->max_edge_score() + model_->node_score( i, c2->label ) ) {
		carpediem_open( x, i-1, c1 );
	    }else {
		return;
	    }
	}
#ifdef LOGGING	    	
	++edge_num_;
#endif
	Real score = c1->alpha + model_->edge_score( c1->label, c2->label, x[i-1] ) + model_->node_score( i, c2->label );
	if ( c1 == first_[i-1] || c2->alpha < score ) {
	    c2->alpha = score;
	    c2->backlink = c1;
	}
    }
}


/*
 * Initialize DP chart for CarpeDiem
 */
template<class Model>
inline void Decoder<Model>::carpediem_init( unsigned int n ) {

    Cell* c = &chart_[0];
    for ( unsigned int i = 0; i < n; ++i ) {
	// Sort label according to node score
	buff_.clear();
	for ( Label label = 0; label < L_; ++label ) {
	    buff_.push_back( std::pair<Label,Real>( label, model_->node_score( i, label ) ) );
	}
	std::sort( buff_.begin(), buff_.end(), cmp_second<Label,Real>() );

	// Initialize DP chart
	for ( unsigned int j = 0; j < L_; ++j ) {
	    c->label    = buff_[j].first;
	    c->alpha = MIN_VALUE;   // use to check open/close
	    if ( j == L_-1 ) {
		c->next  = NULL;
		last_[i] = c;
	    }else {
		c->next = c+1;
	    }
	    if ( j == 0 ) {
		c->prev   = NULL;
		first_[i] = c;
	    }else {
		c->prev = c-1;
	    }
	    ++c;
	}
    }
}




template<class Model>
inline Real Decoder<Model>::fw_staggered( const Input* x, Label* y, unsigned int n, bool prune_flag ) {

    for ( Cell* c = first_[0]; c != NULL; c = c->next ) {
	//for ( Cell* c = chart2_.first(0); c != NULL; c = c->next ) {
	c->alpha    = model_->node_score( 0, c->label );
	c->lb       = c->alpha;
	c->backlink = NULL;
    }
    if ( prune_flag ) prune_chart( 0 );
    //if ( prune_flag ) chart2_.prune( 0, lb_ );
    
    for ( unsigned int i = 1; i < n; ++i ) {
	Cell* c1 = first_[i-1];
	//Cell* c1 = chart2_.first(i-1);
	if ( c1->label < L_ ) {
	    for ( Cell* c2 = first_[i]; c2 != NULL; c2 = c2->next ) {
		//for ( Cell* c2 = chart2_.first(i); c2 != NULL; c2 = c2->next ) {
		Real delta   = model_->edge_score( c1->label, c2->label, x[i-1] ) + model_->node_score( i, c2->label );
		c2->alpha    = c1->alpha + delta;
		c2->lb       = c1->lb + delta;
		c2->backlink = c1;
	    }
	}else {
	    for ( Cell* c2 = first_[i]; c2 != NULL; c2 = c2->next ) {
		//for ( Cell* c2 = chart2_.first(i); c2 != NULL; c2 = c2->next ) {
		c2->alpha    = c1->alpha + model_->edge_score( c1->label, c2->label, x[i-1] ) + model_->node_score( i, c2->label );
		c2->lb       = MIN_VALUE;
		c2->backlink = c1;
	    }
	}
	for ( c1 = c1->next; c1 != NULL; c1 = c1->next ) {
	    if ( c1->label < L_ ) {
		for ( Cell* c2 = first_[i]; c2 != NULL; c2 = c2->next ) {
		    //for ( Cell* c2 = chart2_.first(i); c2 != NULL; c2 = c2->next ) {
		    Real delta = model_->edge_score( c1->label, c2->label, x[i-1] ) + model_->node_score( i, c2->label );
		    Real score = c1->alpha + delta;
		    Real lb    = c1->lb + delta;
		    if ( c2->alpha < score ) {
			c2->alpha = score;
			c2->backlink = c1;
		    }
		    if ( c2->lb < lb ) {
			c2->lb = lb;
		    }
		}
	    }else {
		for ( Cell* c2 = first_[i]; c2 != NULL; c2 = c2->next ) {
		    //for ( Cell* c2 = chart2_.first(i); c2 != NULL; c2 = c2->next ) {
		    Real score = c1->alpha + model_->edge_score( c1->label, c2->label, x[i-1] ) + model_->node_score( i, c2->label );
		    if ( c2->alpha < score ) {
			c2->alpha = score;
			c2->backlink = c1;
		    }
		}
	    }
	}
	if ( prune_flag ) prune_chart( i );
	//if ( prune_flag ) chart2_.prune( i, lb_ );
    }

    // Update lower bound
    if ( prune_flag ) {
	for ( const Cell* c = first_[n-1]; c != NULL; c = c->next ) {
	    //for ( const Cell* c = chart2_.first(n-1); c != NULL; c = c->next ) {
	    if ( c->label < L_ && lb_ < c->lb ) {
		lb_ = c->lb;
	    }
	}
    }
    return traverse_backlink1( y, n );
    //return chart2_.traverse_backlink1( y, n );
}

template<class Model>
inline Real Decoder<Model>::traverse_backlink1( Label* y, unsigned int n ) const {

    const Cell* best_c = NULL;
    Real best_score = MIN_VALUE;
    
    for ( const Cell* c = first_[n-1]; c != NULL; c = c->next ) {
	Real score = c->alpha;
	if ( best_score < score ) {
	    best_c = c;
	    best_score = score;
	}
    }
            
    y[n-1] = best_c->label;
    for ( unsigned int i = n-1; 1 <= i; --i ) {
	best_c = best_c->backlink;
	y[i-1] = best_c->label;
    }
    return best_score;
}


template<class Model>
inline Real Decoder<Model>::bw_staggered( const Input* x, Label* y, unsigned int n, bool prune_flag ) {

    for ( Cell* c = first_[n-1]; c != NULL; c = c->next ) {
    //for ( Cell* c = chart2_.first(n-1); c != NULL; c = c->next ) {
	c->beta     = ZERO;
	c->lb       = ZERO;
	c->backlink = NULL;
    }
    if ( prune_flag ) prune_chart( n-1 );
    //if ( prune_flag ) chart2_.prune( n-1, lb_ );

    for ( unsigned int i = n-1; 1 <= i; --i ) {
	Cell* c2 = first_[i];
	//Cell* c2 = chart2_.first(i);
	if( c2->label < L_ ) {
	    for ( Cell* c1 = first_[i-1]; c1 != NULL; c1 = c1->next ) {
	    //for ( Cell* c1 = chart2_.first(i-1); c1 != NULL; c1 = c1->next ) {
		Real delta   = model_->edge_score( c1->label, c2->label, x[i-1] ) + model_->node_score( i, c2->label );
		c1->beta     = c2->beta + delta;
		c1->lb       = c2->lb + delta;
		c1->backlink = c2;
	    }
	}else {
	    for ( Cell* c1 = first_[i-1]; c1 != NULL; c1 = c1->next ) {
	    //for ( Cell* c1 = chart2_.first(i-1); c1 != NULL; c1 = c1->next ) {
		c1->beta     = c2->beta + model_->edge_score( c1->label, c2->label, x[i-1] ) + model_->node_score( i, c2->label );
		c1->lb       = MIN_VALUE;
		c1->backlink = c2;
	    }
	}
	for ( c2 = c2->next; c2 != NULL; c2 = c2->next ) {
	    if ( c2->label < L_ ) {
		for ( Cell* c1 = first_[i-1]; c1 != NULL; c1 = c1->next ) {
		//for ( Cell* c1 = chart2_.first(i-1); c1 != NULL; c1 = c1->next ) {
		    Real delta = model_->edge_score( c1->label, c2->label, x[i-1] ) + model_->node_score( i, c2->label );
		    Real score = delta + c2->beta;
		    Real lb    = delta + c2->lb;
		    if ( c1->beta < score ) {
			c1->beta = score;
			c1->backlink = c2;
		    }
		    if ( c1->lb < lb ) {
			c1->lb = lb;
		    }
		}
	    }else {
		for ( Cell* c1 = first_[i-1]; c1 != NULL; c1 = c1->next ) {
		//for ( Cell* c1 = chart2_.first(i-1); c1 != NULL; c1 = c1->next ) {
		    Real score = model_->edge_score( c1->label, c2->label, x[i-1] ) + model_->node_score( i, c2->label ) + c2->beta;
		    if ( c1->beta < score ) {
			c1->beta = score;
			c1->backlink = c2;
		    }
		}
	    }
	}
	if ( prune_flag ) prune_chart(i-1);
	//if ( prune_flag ) chart2_.prune(i-1,lb_);
    }

    // Update lower bound
    if ( prune_flag ) {
	for ( const Cell* c = first_[0]; c != NULL; c = c->next ) {
	    //for ( const Cell* c = chart2_.first(0); c != NULL; c = c->next ) {
	    if ( c->label < L_ ) {
		Real score = model_->node_score( 0, c->label ) + c->lb;
		if ( lb_ < score ) {
		    lb_ = score;
		}
	    }
	}
    }
    return traverse_backlink2( y, n );
    //return chart2_.traverse_backlink2( y, n, model_ );
}


/*
 *  Back-tracking for the backward Viterbi algorithm
 */
template<class Model>
inline Real Decoder<Model>::traverse_backlink2( Label* y, unsigned int n ) const {

    const Cell* best_c = NULL;
    Real best_score   = MIN_VALUE;
    
    for ( const Cell* c = first_[0]; c != NULL; c = c->next ) {
	Real score = model_->node_score( 0, c->label ) + c->beta;
	if ( best_score < score ) {
	    best_c = c;
	    best_score = score;
	}
    }
    
    for ( unsigned int i = 0; i < n; ++i ) {
	y[i]   = best_c->label;
	best_c = best_c->backlink;
    }
    return best_score;
}



template<class Model>
inline Real Decoder<Model>::beam( const Input* x, Label* y, unsigned int n ) {

    if ( n == 0 ) return ZERO;

    model_->cache_node_score( x, n );
    viterbi_init( n );

    agenda_init();
    for ( Cell* c = first_[0]; c != NULL; c = c->next ) {
	c->alpha = model_->node_score( 0, c->label );
	c->backlink = NULL;
	agenda_add( c->alpha );
    }
    prune_chart( 0, A_.top() );
    
    for ( unsigned int i = 1; i < n; ++i ) {
	Cell* c1 = first_[i-1];
	for ( Cell* c2 = first_[i]; c2 != NULL; c2 = c2->next ) {
	    c2->alpha = c1->alpha + model_->edge_score( c1->label, c2->label, x[i-1] ) + model_->node_score( i, c2->label );
	    c2->backlink = c1;
	}
	
	for ( c1 = c1->next; c1 != NULL; c1 = c1->next ) {
	    for ( Cell* c2 = first_[i]; c2 != NULL; c2 = c2->next ) {
		Real score = c1->alpha + model_->edge_score( c1->label, c2->label, x[i-1] ) + model_->node_score( i, c2->label );
		if ( c2->alpha < score ) {
		    c2->alpha = score;
		    c2->backlink = c1;
		}
	    }
	}
	
	agenda_init();
	for ( Cell* c2 = first_[i]; c2 != NULL; c2 = c2->next ) {
	    agenda_add( c2->alpha );
	}
	prune_chart( i, A_.top() );
    }

    return traverse_backlink1( y, n );
}


template<class Model>
inline void Decoder<Model>::agenda_add( Real score ) {

    if ( option_.beam_width == A_.size() ) {
	if ( A_.top() < score ) {
	    A_.pop();
	    A_.push( score );
	}
    }else {
	A_.push( score );
    }
}


template<class Model>
inline void Decoder<Model>::agenda_init() {

    while ( A_.empty() == false ) {
	A_.pop();
    }
}


/*
 * Deterministic decoding
 */
template<class Model>
inline Real Decoder<Model>::det( const Input* x, Label* y, unsigned int n ) {

    if ( n == 0 ) return ZERO;

    model_->cache_node_score( x, n );
    Real best_score = MIN_VALUE;
    for ( Label label = 0; label < L_; ++label ) {
	Real score = model_->node_score( 0, label );
	if ( best_score < score ) {
	    y[0] = label;
	    best_score = score;
	}
    }
    for ( unsigned int i = 1; i < n; ++i ) {
	Real best_delta = MIN_VALUE;
	for ( Label label = 0; label < L_; ++label ) {
	    Real delta = model_->edge_score( y[i-1], label, x[i-1] ) + model_->node_score( i, label );
	    if ( best_delta < delta ) {
		y[i] = label;
		best_delta = delta;
	    }
	}
	best_score += best_delta;
    }
    return best_score;
}


/*
 * Staggered decoding
 */
template<class Model>
inline Real Decoder<Model>::staggered( const Input* x, Label* y, unsigned int n ) {

    if ( n == 0 ) return ZERO;

    timer1_->startTimer();
    if ( option_.initialization == NEG_INF ) {
	lb_ = MIN_VALUE;
	model_->cache_node_score( x, n );
    }else {
	lb_ = det( x, y, n );
    }
    staggered_init( n );
    //chart2_.staggered_init( n );
    timer1_->stopTimer();

    timer2_->startTimer();
    // Iteration
    Real score = MIN_VALUE;
    for ( iter_num_ = 0; ; ++iter_num_ ) {
	if ( iter_num_%2 == 0 ) {
	    score = fw_staggered( x, y, n, option_.pruning&&(iter_num_!=0));
	}else {
	    score = bw_staggered( x, y, n, option_.pruning );
	}
	bool is_optimal = true;
	for ( unsigned int i = 0; i < n; ++i ) {
	    if ( L_ <= y[i] ) {
		is_optimal = false;
		break;
	    }
	}

	if ( is_optimal ) {
	    timer2_->stopTimer();
	    return score;
	}else if ( option_.expansion == FLAT ) {
	    for ( unsigned int i = 0; i < n; ++i ) {
		if ( L_ <= last_[i]->label ) {
		    expand_chart( i );
		    //chart2_.expand( i, model_ );
		}
	    }
	}else {
	    for ( unsigned int i = 0; i < n; ++i ) {
		if ( L_ <= y[i] ) {
		    expand_chart( i );
		    //chart2_.expand( i, model_ );
		}
	    }
	}
    }
}


// Prune cells for the beam search
template<class Model>
inline void Decoder<Model>::prune_chart( unsigned int i, Real th ) {

    for ( Cell* c = first_[i]; c != NULL; c = c->next ) {
	if ( c->alpha < th ) {
	    Cell* c_prev = c->prev;
	    Cell* c_next = c->next;
	    if ( c_prev == NULL ) {
		c_next->prev = NULL;
		first_[i] = c_next;
	    }else if ( c_next == NULL ) {
		c_prev->next = NULL;
		last_[i] = c_prev;
	    }else {
		c_prev->next = c_next;
		c_next->prev = c_prev;
	    }
	}
    }
}


/*
 * Prune cells in i-th column
 */
template<class Model>
inline void Decoder<Model>::prune_chart( unsigned int i ) {

    for ( Cell* c = first_[i]; c != NULL; c = c->next ) {
	if ( c->alpha + c->beta + EPSILON < lb_ ) {
	    Cell* c_prev = c->prev;
	    Cell* c_next = c->next;
	    if ( c_prev == NULL ) {
		if ( c_next == NULL ) {
		    fprintf( stderr, "Pruining error\n" );
		    exit( -1 );
		}
		c_next->prev = NULL;
		first_[i] = c_next;
	    }else if ( c_next == NULL ) {
		c_prev->next = NULL;
		last_[i] = c_prev;
	    }else {
		c_prev->next = c_next;
		c_next->prev = c_prev;
	    }
	}
    }
}


/*
 * Expand i-th column
 */
template<class Model>
inline void Decoder<Model>::expand_chart( unsigned int i ) {

    // Memorize the last cell (= degenerate label)
    Cell* last     = last_[i];
    Label label    = last->label;
    Real  alpha = last->alpha;
    Real  beta = last->beta;
    Cell* prev     = last->prev;
	
    // Expand active labels
    if ( label+1 < L_+D_ ) {
	Cell* c = last;
	Label y_first = pow2( label-L_ );
	Label y_last = 2*y_first;
	for ( Label y = y_first; y < y_last; ++y ) {
	    c->label    = y;
	    c->alpha = alpha + model_->node_score( i, y ) - model_->node_score( i, label );
	    c->beta = beta;
	    if ( y == y_first ) {
		c->prev = prev;
	    }else {
		c->prev = c - 1;
	    }
	    c->next = c + 1;
	    ++c;
	}
	
	// degenerate label
	c->label    = label + 1;
	c->alpha = alpha + model_->node_score( i, c->label ) - model_->node_score( i, label );
	c->beta = beta;
	c->next     = NULL;
	c->prev     = c - 1;
	last_[i]    = c;
    }else {
	Cell* c = last;
	Label y_first = pow2( label-L_ );
	for ( Label y = y_first; y < L_; ++y ) {
	    c->label    = y;
	    c->alpha = alpha + model_->node_score( i, y ) - model_->node_score( i, label );
	    c->beta = beta;
	    if ( y == y_first ) {
		c->prev = prev;
	    }else {
		c->prev = c - 1;
	    }
	    if ( y == L_ - 1 ) {
		c->next = NULL;
		last_[i] = c;
	    }else {
		c->next = c + 1;
		++c;
	    }
	}
    }
}


/*
 * Obtain the chart information for debugging.
 */
template<class Model>
inline void Decoder<Model>::debug_chart( unsigned int i, std::vector<Label>& y ) const {

    y.clear();
    for ( const Cell* c = first_[i]; c != NULL; c = c->next ) {
	y.push_back( c->label );
    }
}


template<class Model>
inline unsigned int Decoder<Model>::iter_num() const {
    
    return iter_num_+1;
}


template<class Model>
inline float Decoder<Model>::edge_rate() const {
    
    return edge_rate_;
}



/*
 * Initialization for the Viterbi algorithm
 */
template<class Model>
inline void Decoder<Model>::viterbi_init( unsigned int n ) {

    Cell* c = &chart_[0];
    for ( unsigned int i = 0; i < n; ++i ) {
	for ( Label y = 0; y < L_; ++y ) {
	    c->label = y;
	    if ( y == 0 ) {
		c->prev   = NULL;
		first_[i] = c;
	    }else {
		c->prev = c-1;
	    }
	    if ( y == L_-1 ) {
		c->next  = NULL;
		last_[i] = c;
	    }else {
		c->next = c+1;
	    }
	    ++c;
	}
    }
}


/*
 *  Initialize for staggered decoding
 */
template<class Model>
inline void Decoder<Model>::staggered_init( unsigned int n ) {

    Cell* c = &chart_[0];
    for ( unsigned int i = 0; i < n; ++i ) {
	// active label
	c->label  = 0;
	c->next   = c + 1;
	c->prev   = NULL;
	first_[i] = c;
	++c;
	
	// degenerate label
	c->label = L_;
	c->next  = NULL;
	c->prev  = c - 1;
	last_[i] = c;
	c += (L_-1);
    }
}


}

#endif

