#ifndef __SD_DenseMatrix__
#define __SD_DenseMatrix__

#include <iostream>
#include <fstream>

#include "util.h"
#include "dense_vector.h"

namespace SD {
    
class DenseMatrix {
  private:
    unsigned int row_size_;
    unsigned int column_size_;
    std::vector<DenseVector> data_;
  public:
    DenseMatrix();
    DenseMatrix( const DenseMatrix& obj );
    ~DenseMatrix() {};
    DenseMatrix& operator=( const DenseMatrix& obj );
    const DenseVector& operator[]( unsigned int row ) const;
    DenseVector& operator[]( unsigned int row );
    DenseMatrix& operator/=( Real v );
    DenseMatrix& operator-=( const DenseMatrix& obj );
    int load( FILE* is );
    int save( FILE* os ) const;
    void init( unsigned int row_size, unsigned int column_size, Real init_val=0.0 );
    unsigned int row_size() const;
    unsigned int column_size() const;
    Real max() const;
    Real max( unsigned int first_row, unsigned int last_row, unsigned int first_column, unsigned int last_column ) const;
    void fill( Real val );
};


inline DenseMatrix::DenseMatrix() {

    row_size_    = 0;
    column_size_ = 0;
}


inline DenseMatrix::DenseMatrix( const DenseMatrix& obj ) {

    row_size_    = obj.row_size();
    column_size_ = obj.column_size();
    data_.resize( row_size_ );
    for ( unsigned int i = 0; i < row_size_; ++i ) {
	data_[i] = obj[i];
    }
}


inline DenseMatrix& DenseMatrix::operator=( const DenseMatrix& obj ) {

    if ( this != &obj ) {
	row_size_    = obj.row_size();
	column_size_ = obj.column_size();
	data_.resize( row_size_ );
	for ( unsigned int i = 0; i < row_size_; ++i ) {
	    data_[i] = obj[i];
	}
    }
    return *this;
}


inline DenseMatrix& DenseMatrix::operator/=( Real v ) {

    for ( unsigned int i = 0; i < row_size_; ++i ) {
	data_[i] /= v;
    }
    return *this;
}


inline DenseMatrix& DenseMatrix::operator-=( const DenseMatrix& obj ) {

    if ( this == &obj ) {
	for ( unsigned int i = 0; i < row_size_; ++i ) {
	    data_[i].fill( ZERO );
	}
    }else {
	for ( unsigned int i = 0; i < row_size_; ++i ) {
	    data_[i] -= obj[i];
	}
    }
    return *this;
}
    
inline int DenseMatrix::load( FILE* is ) {

    if ( fread( &row_size_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( fread( &column_size_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    data_.resize( row_size_ );
    for ( unsigned int i = 0; i < row_size_; ++i ) {
	if ( data_[i].load( is ) == FAILURE ) {
	    return FAILURE;
	}
    }
    return SUCCESS;
}


inline int DenseMatrix::save( FILE* os ) const {

    if ( fwrite( &row_size_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &column_size_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    for ( unsigned int i = 0; i < row_size_; ++i ) {
	if ( data_[i].save( os ) == FAILURE ) {
	    return FAILURE;
	}
    }
    return SUCCESS;
}


inline unsigned DenseMatrix::row_size() const {

    return row_size_;
}


inline unsigned int DenseMatrix::column_size() const {

    return column_size_;
}


inline void DenseMatrix::init( unsigned int row_size, unsigned int column_size, Real val ) {
    
    row_size_    = row_size;
    column_size_ = column_size;
    data_.resize( row_size_ );
    for ( unsigned int i = 0; i < row_size_; ++i ) {
	data_[i].init( column_size_, val );
    }
}


inline const DenseVector& DenseMatrix::operator[]( unsigned int row ) const {

    return data_[row];
}


inline DenseVector& DenseMatrix::operator[]( unsigned int row ) {

    return data_[row];
}


inline Real DenseMatrix::max() const {
    
    return this->max( 0, row_size_, 0, column_size_ );
}


inline Real DenseMatrix::max( unsigned int first_row, unsigned int last_row, unsigned int first_column, unsigned int last_column ) const {

    Real val = data_[first_row][first_column];
    for ( unsigned int row = first_row; row < last_row; ++row ) {
	val = std::max<Real>( val, data_[row].max( first_column, last_column ) );
    }
    return val;
}

inline void DenseMatrix::fill( Real val ) {

    for ( unsigned int i = 0; i < row_size_; ++i ) {	
	data_[i].fill( val );
    }
}

}

#endif
