#ifndef __SD_DenseVector__
#define __SD_DenseVector__

#include <iostream>
#include <fstream>
#include <algorithm> // max, max_element, copy, back_inserter
#include <numeric>   // accumulate
#include <math.h>   // log
#include <vector>
#include <float.h>

#include "util.h"
#include "sparse_vector.h"

namespace SD {
    
class DenseVector {
  private:
    unsigned int size_;
    std::vector<Real> data_;
  public:
    DenseVector();
    DenseVector( const DenseVector& obj );
    ~DenseVector() {};
    DenseVector& operator=( const DenseVector& obj );
    DenseVector& operator/=( Real val );
    DenseVector& operator-=( const DenseVector& obj );
    DenseVector& operator+=( const SparseVector& sv );
    int load( FILE* is );
    int save( FILE* os ) const;
    void init( unsigned int size, Real val );
    unsigned int size() const;
    Real operator[]( unsigned int i ) const;
    Real& operator[]( unsigned int i );
    Real max( unsigned int first, unsigned int last ) const;
    Real sum( unsigned int first, unsigned int last ) const;
    void fill( Real val );
    void estimate_ub( unsigned int L, unsigned int D, unsigned int M );
    void logarithmize();
    std::vector<Real>::const_iterator begin() const;
    std::vector<Real>::const_iterator end() const;
};


inline DenseVector::DenseVector() {

    size_ = 0;
}


inline DenseVector::DenseVector( const DenseVector& obj ) {

    size_ = obj.size();
    std::copy( obj.begin(), obj.end(), std::back_inserter( data_ ) );
}


inline DenseVector& DenseVector::operator=( const DenseVector& obj ) {

    if ( this != &obj ) {
	size_ = obj.size();
	data_.clear();
	std::copy( obj.begin(), obj.end(), std::back_inserter( data_ ) );
    }
    return *this;
}


inline DenseVector& DenseVector::operator/=( Real val ) {

    for ( std::vector<Real>::iterator it = data_.begin(); it != data_.end(); ++it ) {
	*it /= val;
    }
    return *this;
}


inline DenseVector& DenseVector::operator-=( const DenseVector& obj ) {

    if ( this == &obj ) {
	std::fill( data_.begin(),data_.end(), ZERO );
    }else {
	for ( unsigned int i = 0; i < size_; ++i ) {
	    data_[i] -= obj[i];
	}
    }
    return *this;
}


inline DenseVector& DenseVector::operator+=( const SparseVector& sv ) {

    for ( SparseVector::const_iterator it = sv.begin(); it != sv.end(); ++it ) {
	data_[it->first] += it->second;
    }
    return *this;
}



inline int DenseVector::load( FILE* is ) {

    if ( fread( &size_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    data_.resize( size_ );
    if ( fread( &data_[0], sizeof( Real ), size_, is ) != size_ ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    return SUCCESS;
}


inline int DenseVector::save( FILE* os ) const {

    if ( fwrite( &size_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &data_[0], sizeof( Real ), size_, os ) != size_ ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    return SUCCESS;
}


inline unsigned DenseVector::size() const {

    return size_;
}


inline void DenseVector::init( unsigned int size, Real val ) {

    size_ = size;
    data_.clear();
    data_.resize( size_, val );
}


inline Real DenseVector::operator[]( unsigned int i ) const {

    return data_[i];
}


inline Real& DenseVector::operator[]( unsigned int i ) {

    return data_[i];
}


inline Real DenseVector::max( unsigned int first, unsigned int last ) const {

    return *std::max_element( &data_[0]+first, &data_[0]+last );
}


inline Real DenseVector::sum( unsigned int first, unsigned int last ) const {

    return std::accumulate( &data_[0]+first, &data_[0]+last, ZERO );
}


inline void DenseVector::fill( Real val ) {

    std::fill( data_.begin(), data_.end(), val );
}


//
//  Use this method to estimate parameters of the degenerate lattice.
//
//  arguments:
//  L: # of original labels
//  D: # of degenerate labels
//  A: max. # of active labels = pow2(D-1)
//
//  data_[0]..data_[L-1]: parameters of the orignal labels (given)
//  data_[L]..data_[L+D-1]: parameters of the degenerate labels (to be estimated)
//
inline void DenseVector::estimate_ub( unsigned int L, unsigned int D, unsigned int A ) {

    data_[L+D-1] = this->max( A, L );
    for ( unsigned int z = L+D-2; L <= z; --z ) {
	A /= 2;
	data_[z] = std::max<Real>( this->max( A, 2*A ), data_[z+1] );
    }
}


inline void DenseVector::logarithmize() {

    for ( std::vector<Real>::iterator it = data_.begin(); it != data_.end(); ++it ) {
	*it = log(*it);
    }
}


inline std::vector<Real>::const_iterator DenseVector::begin() const {

    return data_.begin();
}


inline std::vector<Real>::const_iterator DenseVector::end() const {

    return data_.end();
}


}

#endif
