#ifndef __DOUBLEARRAY__
#define __DOUBLEARRAY__

#include <iostream>
#include <fstream>
#include <algorithm> // sort, copy
#include <string>
#include <vector>
#include <set>
#include "util.h"

/*
 * Simple static doulbe array
 */
namespace SD {

class DoubleArray {
  private:
    typedef unsigned int uint;
    typedef unsigned char uchar;
    static const char ENDMARKER;
    static const char DUMMYCHAR;
    static const unsigned int ROOT;

    // the double array
    unsigned int              array_size_;
    std::vector<unsigned int> base_;
    std::vector<char>         check_;
        
    // key and key_id
    unsigned int              key_size_;
    std::vector<std::string>  key_;
    std::vector<unsigned int> key_id_;    // mapping from base value to key ID
  public:
    DoubleArray() {};
    ~DoubleArray() {};
    DoubleArray( const DoubleArray& rhs );
    DoubleArray& operator=( const DoubleArray& rhs );
    
    int build( const std::vector<std::string>& key );
    int save( FILE* os ) const;
    int load( FILE* is );
    void read( const char* s, const char* last, uint node_attr_num, std::vector<uint>& node_id, std::vector<uint>& edge_id ) const;
    
    const char* key( uint key_id ) const;
    unsigned int key_id( const std::string& key ) const;
    unsigned int key_id( const char* key ) const;
    unsigned int key_size() const;
  private:
    void init_array();
    void build( unsigned int n, const std::string& prefix, const std::vector<std::string>& v, unsigned int& min_base, std::set<unsigned int>& reserved_id );
    int store_key( const std::vector<std::string>& key );
    void decide_base( unsigned int n, unsigned int min_base, const std::set<char>& c, const std::set<unsigned int>& reserved_id );
};

//  Use control characters as the special characters.
const char DoubleArray::ENDMARKER = '\x00';
const char DoubleArray::DUMMYCHAR = '\x01';
const unsigned int DoubleArray::ROOT = 1;


/**********************************************************************
 *
 *  NOTE:
 *
 *  Every node ID (including the root) is equal or larger than 1.
 *
 *  const ROOT(=1) is the ID of the root node.
 *
 *  Min value of base is 1. This gurantees that base_[i] + c => 1
 *  for all i and c.
 *
 *  check_[i] = DUMMYCHAR implies that node ID i does not exist.
 *
 **********************************************************************/

inline int DoubleArray::build( const std::vector<std::string>& key ) {

    // initialize the double array
    init_array();

    // copy the argument, and append ENDMARKER to the keys.
    std::vector<std::string> v;
    std::copy( key.begin(), key.end(), std::back_inserter(v) );
    for ( uint i = 0; i < key.size(); i++ ) {
	v[i].append( 1, ENDMARKER );
    }
    std::sort( v.begin(), v.end() );

    // the main part
    uint min_base = 1;
    std::set<unsigned int> reserved_id;
    reserved_id.insert( ROOT );
    build( ROOT, "", v, min_base, reserved_id );

    if ( store_key( key ) == FAILURE ) return FAILURE;
    
    return SUCCESS;
}


inline void DoubleArray::init_array() {
    
    array_size_ = 2;
    
    base_.resize( array_size_ );
    base_[0] = 0;
    base_[1] = 1;
    
    check_.resize( array_size_ );
    check_[0] = DUMMYCHAR;
    check_[1] = DUMMYCHAR;
}


inline int DoubleArray::store_key( const std::vector<std::string>& key ) {

    key_size_ = key.size();
    std::copy( key.begin(), key.end(), std::back_inserter(key_) );

    key_id_.resize( array_size_, key_size_ );
    for ( unsigned int i = 0; i < key_size_; ++i ) {
	unsigned int node = ROOT;
	for ( const char* s = key_[i].c_str(); *s != '\0'; ++s ) {
	    unsigned int next = base_[node] + static_cast<unsigned int>(static_cast<unsigned char>(*s));
	    if ( array_size_ < next || check_[next] != *s ) {
		fprintf( stderr, HERE "unexpected error\n" );
		return FAILURE;
	    }
	    node = next;
	}
	unsigned int next = base_[node] + static_cast<unsigned int>(static_cast<unsigned char>(ENDMARKER));
	if ( array_size_ < next || check_[next] != ENDMARKER ) {
	    fprintf( stderr, HERE "unexpected error\n" );
	    return FAILURE;
	}
	key_id_[next] = i;
    }
    
    return SUCCESS;
}


inline void DoubleArray::build( uint n, const std::string& prefix, const std::vector<std::string>& v, uint& min_base, std::set<uint>& reserved_id ) {

    // prefix ³Ƥʸ󤹤
    std::set<char> successor;
    std::vector<std::string>::const_iterator lb = lower_bound( v.begin(), v.end(), prefix );
    while ( lb != v.end() && lb->find( prefix, 0 ) == 0 ) {
	if ( *lb != prefix ) {
	    successor.insert( lb->at( prefix.length() ) );
	}
	++lb;
    }
    if ( successor.size() == 0 ) return;

    // decide base_[n]
    decide_base( n, min_base, successor, reserved_id );
	
    //  arrange the array size
    uint max = 0;
    for ( std::set<char>::iterator it = successor.begin(); it != successor.end(); ++it ) {
	if ( max < base_[n] + static_cast<uint>(static_cast<uchar>(*it)) ) {
	    max = base_[n] + static_cast<uint>(static_cast<uchar>(*it));
	}
    }
    if ( array_size_ < max + 1 ) {
	base_.resize( max+1 );
	check_.resize( max+1, DUMMYCHAR );
	array_size_ = max+1;
    }
    
    //  set the check_
    for ( std::set<char>::iterator it = successor.begin(); it != successor.end(); ++it ) {
	check_[base_[n] + static_cast<uint>(static_cast<uchar>(*it))] = *it;
    }
    
    //  update min_base and reserved_id
    min_base = base_[n] + 1;
    for ( std::set<char>::iterator it = successor.begin(); it != successor.end(); ++it ) {
	reserved_id.insert( base_[n] + static_cast<uint>(static_cast<uchar>(*it)) );
    }
    
    //  process recursively
    for ( std::set<char>::iterator it = successor.begin(); it != successor.end(); ++it ) {
	build( base_[n] + static_cast<uint>(static_cast<uchar>(*it)), prefix + *it, v, min_base, reserved_id );
    }
}


void DoubleArray::decide_base( uint n, uint min_base, const std::set<char>& c, const std::set<uint>& reserved_id ) {

    base_[n] = min_base;
    bool collision = false;
    do {
	// check collision
	collision = false;
	for ( std::set<char>::const_iterator it = c.begin(); it != c.end(); ++it ) {
	    if ( reserved_id.find( base_[n] + static_cast<uint>(static_cast<uchar>(*it)) ) != reserved_id.end() ) {
		collision = true;
		break;
	    }
	}
	
	//  ͤʤʤޤ array_[n].base ͤ䤷Ƥ
	if ( collision == true ) {
	    base_[n]++;
	}
    }while( collision == true );
}



inline int DoubleArray::save( FILE* os ) const {

    if ( fwrite( &array_size_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &key_size_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &base_[0], sizeof( unsigned int ), array_size_, os ) != array_size_ ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &check_[0], sizeof( char ), array_size_, os ) != array_size_ ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &key_id_[0], sizeof( unsigned int ), array_size_, os ) != array_size_ ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    for ( unsigned int i = 0; i < key_size_; ++i ) {
	unsigned int size = key_[i].length() + 1; // nullʸޤ
	if ( fwrite( &size, sizeof( unsigned int ), 1, os ) != 1 ) {
	    fprintf( stderr, HERE "failed to save\n" );
	    return FAILURE;
	}
	if ( fwrite( key_[i].c_str(), sizeof( char ), size, os ) != size ) {
	    fprintf( stderr, HERE "failed to save\n" );
	    return FAILURE;
	}
    }
    return SUCCESS;
}


inline int DoubleArray::load( FILE* is ) {

    if ( fread( &array_size_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( fread( &key_size_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    base_.resize( array_size_ );
    if ( fread( &base_[0], sizeof( unsigned int ), array_size_, is ) != array_size_ ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    check_.resize( array_size_ );
    if ( fread( &check_[0], sizeof( char ), array_size_, is ) != array_size_ ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    key_id_.resize( array_size_ );
    if ( fread( &key_id_[0], sizeof( unsigned int ), array_size_, is ) != array_size_ ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    key_.resize( key_size_ );
    for ( unsigned int i = 0; i < key_size_; ++i ) {
	unsigned int size;
	if ( fread( &size, sizeof( unsigned int ), 1, is ) != 1 ) {
	    fprintf( stderr, HERE "failed to load\n" );
	    return FAILURE;
	}
	char buff[size];
	if ( fread( buff, sizeof( char ), size, is ) != size ) {
	    fprintf( stderr, HERE "failed to load\n" );
	    return FAILURE;
	}
	key_[i].assign( buff );
    }
    return SUCCESS;
}


inline void DoubleArray::read( const char* s, const char* last, unsigned int node_attr_num, std::vector<unsigned int>& node_id, std::vector<unsigned int>& edge_id ) const {
    
    node_id.clear();
    edge_id.clear();

    bool is_node_attr = true;
    for ( ;s != last; ++s ) {
	bool flag = true;
	unsigned int node = ROOT;

	if ( *s == 'B' ) {
	    is_node_attr = false;
	}
	for ( ; *s != ' ' && *s != '\t'; ++s ) {
	    unsigned int next = base_[node] + static_cast<unsigned int>(static_cast<uchar>(*s));
	    if ( next < array_size_ && check_[next] == *s ) {
		node = next;
	    }else {
		flag = false;
		for ( ; *s != ' ' && *s != '\t'; ++s ) {}
		break;
	    }
	}
	if ( flag == true ) {
	    unsigned int next = base_[node] + static_cast<unsigned int>(static_cast<uchar>(ENDMARKER));
	    if ( next < array_size_ && check_[next] == ENDMARKER ) {
		if ( is_node_attr == true ) {
		    node_id.push_back( key_id_[next] );
		}else {
		    edge_id.push_back( key_id_[next] - node_attr_num );
		}
	    }
	}
    }
}



inline unsigned int DoubleArray::key_id( const std::string& key ) const {

    return key_id( key.c_str() );
}


inline unsigned int DoubleArray::key_id( const char* key ) const {

    unsigned int node = ROOT;
    for ( const char* s = key; *s != '\0'; ++s ) {
	unsigned int next = base_[node] + static_cast<unsigned int>(static_cast<unsigned char>(*s));
	if ( next < array_size_ && check_[next] == *s ) {
	    node = next;	    
	}else {
	    return key_size_;
	}
    }
    
    unsigned int next = base_[node] + static_cast<unsigned int>(static_cast<unsigned char>(ENDMARKER));
    if ( next < array_size_ && check_[next] == ENDMARKER ) {
	return key_id_[next];
    }else {
	return key_size_;
    }
}


inline const char* DoubleArray::key( unsigned int i ) const {

    return key_[i].c_str();
}


inline unsigned int DoubleArray::key_size() const {
    
    return key_size_;
}

}
#endif
