#ifndef __SD_HMM__
#define __SD_HMM__

#include <iostream>
#include <fstream>

#include "dense_vector.h"
#include "dense_matrix.h"
#include "util.h"

namespace SD {

class HMM {
  public:
    typedef unsigned int Input;
  private:
    unsigned int L_;  // # of original labels
    unsigned int D_;  // # of degenerate labels
    unsigned int K_;  // # of all labels (= L_+D_)
    unsigned int A_;  // max. # of active labels (= pow2(D_-1))
    unsigned int V_;  // # of tokens including unknown word

    DenseVector pi_;      // initial state score (= log prob.)
    DenseMatrix a_;       // transition score
    DenseMatrix b_;       // emission score
    DenseMatrix cache_;   // cached node score
    Real        max_a_;
  public:
    HMM() {};
    HMM( const HMM& rhs );
    ~HMM() {};
    HMM& operator=( const HMM& rhs );
    
    // Training
    void init( unsigned int L, unsigned int V );
    void update( const Input* x, const Label* y, unsigned int n );
    void finish();
    
    // Load/save model file
    int load( FILE* is );
    int save( FILE* os ) const;
    
    // Obtain scores
    void cache_node_score( const Input* x, unsigned int n );
    Real node_score( unsigned int i, const Label& y ) const;
    Real edge_score( const Label& y1, const Label& y2, const Input& x ) const;
    Real max_edge_score() const;
};


inline void HMM::init( unsigned int L, unsigned int V ) {

    L_ = L;
    D_ = degenerate_label_num( L_ );
    K_ = L_ + D_;
    A_ = pow2(D_-1);
    V_ = V;
    
    // alpha is a smoothing factor
    Real alpha = 0.1;
    pi_.init( K_, alpha );
    a_.init( K_, K_, alpha );
    b_.init( K_, V_, alpha );
}


inline void HMM::update( const Input* x, const Label* y, unsigned int n ) {
    
    for ( unsigned int i = 0; i < n; ++i ) {
	if ( i == 0 ) {
	    pi_[y[i]] += 1.0;
	}else {
	    a_[y[i-1]][y[i]] += 1.0;
	}
	b_[y[i]][x[i]] += 1.0;
    }
}


inline void HMM::finish() {

    // compute log-probability
    pi_ /= pi_.sum( 0, L_ );
    pi_.logarithmize();
    for ( Label y = 0; y < L_; ++y ) {
	a_[y] /= a_[y].sum( 0, L_ );
	a_[y].logarithmize();
	b_[y] /= b_[y].sum( 0, V_ );
	b_[y].logarithmize();
    }
    
    pi_.estimate_ub( L_, D_, A_ );
    for ( Label y = 0; y < L_; ++y ) {
	for ( Label z = L_; z < K_; ++z ) {
	    a_[y][z] = a_.max( y, y+1, pow2(z-L_), L_ );
	    a_[z][y] = a_.max( pow2(z-L_), L_, y, y+1 );
	}
    }
    for ( Label z1 = L_; z1 < K_; ++z1 ) {
	for ( Label z2 = L_; z2 < K_; ++z2 ) {
	    a_[z1][z2] = a_.max( pow2(z1-L_), L_, pow2(z2-L_), L_ );
	}
    }
    for ( Label z = L_; z < K_; ++z ) {
	for ( Input x = 0; x < V_; ++x ) {
	    b_[z][x] = b_.max( pow2(z-L_), L_, x, x+1 );
	}
    }
    max_a_ = a_.max();
}


inline int HMM::load( FILE* is ) {

    if ( fread( &L_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }    
    if ( fread( &D_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( fread( &K_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( fread( &A_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( fread( &V_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( pi_.load( is ) == FAILURE ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( a_.load( is ) == FAILURE ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( fread( &max_a_, sizeof( Real ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( b_.load( is ) == FAILURE ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    cache_.init( MAX_LENGTH, K_, ZERO );
    
    return SUCCESS;
}


inline int HMM::save( FILE* os ) const {

    if ( fwrite( &L_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &D_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &K_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &A_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &V_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( pi_.save( os ) == FAILURE ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( a_.save( os ) == FAILURE ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &max_a_, sizeof( Real ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( b_.save( os ) == FAILURE ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    return SUCCESS;
}


inline void HMM::cache_node_score( const Input* x, unsigned int n ) {

    for ( unsigned int i = 0; i < n; ++i ) {
	for ( Label y = 0; y < L_; ++y ) {
	    cache_[i][y] = b_[y][x[i]];
	}
	if ( i == 0 ) {
	    for ( Label y = 0; y < L_; ++y ) {
		cache_[i][y] += pi_[y];
	    }
	}
	cache_[i].estimate_ub( L_, D_, A_ );
    }
}


inline Real HMM::node_score( unsigned int i, const Label& y ) const {

    return cache_[i][y];
}

    
inline Real HMM::edge_score( const Label& y1, const Label& y2, const Input& ) const { // x

    return a_[y1][y2];
}


inline Real HMM::max_edge_score() const {
    
    return max_a_;
}


}

#endif
