#ifndef __SD_HMM_TAGGER__
#define __SD_HMM_TAGGER__

#include <iostream>
#include <fstream>
#include <cstring> // strlen

#include "double_array.h"
#include "hmm.h"
#include "decoder.h"
#include "result.h"
#include "test_option.h"
#include "util.h"

namespace SD {

class HMMTagger {
  private:
    typedef HMM::Input Input;
    DoubleArray token_indexer_;
    DoubleArray label_indexer_;
    HMM         hmm_;
  public:
    HMMTagger() {};
    HMMTagger( const HMMTagger& obj );
    ~HMMTagger() {};
    int load( const char* modelfile ) const;
    int test( const char* testfile, const TestOption& option );
};


inline int HMMTagger::test( const char* testfile, const TestOption& option ) {

    Decoder<HMM> decoder;
    decoder.init( label_indexer_.key_size(), degenerate_label_num(label_indexer_.key_size()), &hmm_, option.decoding_option() );
    
    TimerPool pool;
    Timer* timer  = pool.push( "Decoding time" );
    
    Result result;
    unsigned int n = 0;
    Input* x       = new Input[MAX_LENGTH];
    Label* y       = new Label[MAX_LENGTH];
    Label* y_best  = new Label[MAX_LENGTH];
    char* line     = new char[BUFF_SIZE];
    char* token    = new char[BUFF_SIZE];
    char* label    = new char[BUFF_SIZE];
    
    FILE* is = fopen( testfile, "r" );
    if ( is == NULL ) {
	fprintf( stderr, "cannot open %s\n", testfile );
	return FAILRE;
    }
    setvbuf( is, NULL, _IOFBF, 1024*1024 );
    while ( fgets( line, BUFF_SIZE, is ) != NULL) {
	if ( line[strlen(line)-1] != '\n' ) {
	    fprintf( stderr, HERE "buffer overflow\n" );
	    return FAILRE;
	}
	line[strlen(line)-1] = '\0';

	if ( strlen(line) == 0 ) {
	    timer->startTimer();
	    Real score = decoder.run( x, y_best, n );
	    timer->stopTimer();
	    result.update( y, y_best, n, decoder );
	    
	    if ( option.verbose() == 1 ) {
		printf( "# log-likelihood = %f\n", score );
		for ( unsigned int i = 0; i < n; ++i ) {
		    printf( "%s\n", label_indexer_.key(y_best[i]) );
		}
		printf( "\n" );
	    }
	    n = 0;
	}else if ( sscanf( line, "%s %s", token, label ) == 2 ) {
	    x[n] = token_indexer_.key_id(token);
	    y[n] = label_indexer_.key_id(label);
	    ++n;
	}else {
	    fprintf( stderr, "Invalid format: %s\n", line );
	    return FAILRE;
	}
    }
    fclose( is );
    
    delete[] line;
    delete[] token;
    delete[] label;
    delete[] x;
    delete[] y;
    delete[] y_best;

    if ( option.verbose() == 0 ) {
	pool.print();
	result.print();
    }
    
    return SUCCESS;
}


inline int HMMTagger::load( const char* modelfile ) {

    FILE* is = fopen( modelfile, "rb" );
    if ( is == NULL ) {
	fprintf( stderr, "cannot open %s\n", modelfile );
	return FAILURE;
    }
    if ( token_indexer_.load(is) == FAILURE ) return FAILURE;
    if ( label_indexer_.load(is) == FAILURE ) return FAILURE;
    if ( hmm_.load(is) == FAILURE ) return FAILURE;
    fclose( is );
    
    return SUCCESS;
}


}

#endif

