#include <iostream>
#include <fstream>

#include "hmm.h"
#include "decoder.h"
#include "double_array.h"
#include "test_option.h"
#include "result.h"
#include "timer.h"

using namespace SD;

int test( const char* testfile, const DoubleArray& token_indexer, const DoubleArray& label_indexer, Decoder<HMM>& decoder, unsigned int verbose ) {

    TimerPool pool;
    Timer* timer  = pool.push( "Decoding time" );

    unsigned int n = 0;
    Result result;
    HMM::Input* x      = new HMM::Input[MAX_LENGTH];
    Label* y      = new Label[MAX_LENGTH];
    Label* y_best = new Label[MAX_LENGTH];
    char* line         = new char[MAX_BYTE_LENGTH];
    char* token        = new char[MAX_BYTE_LENGTH];
    char* label        = new char[MAX_BYTE_LENGTH];
    
    FILE* is = fopen( testfile, "r" );
    if ( is == NULL ) {
	fprintf( stderr, "cannot open %s\n", testfile );
	return -1;
    }
    setvbuf( is, NULL, _IOFBF, 1024*1024 );
    while ( fgets( line, MAX_BYTE_LENGTH, is ) != NULL) {
	if ( line[strlen(line)-1] != '\n' ) {
	    fprintf( stderr, HERE "buffer overflow\n" );
	    return -1;
	}
	line[strlen(line)-1] = '\0';

	if ( strlen(line) == 0 ) {
	    timer->startTimer();
	    Real score = decoder.run( x, y_best, n );
	    timer->stopTimer();
	    result.update( y, y_best, n, decoder );
	    
	    if ( verbose == 1 ) {
		printf( "# log-likelihood = %f\n", score );
		for ( unsigned int i = 0; i < n; ++i ) {
		    printf( "%s\n", label_indexer.key(y_best[i]) );
		}
		printf( "\n" );
	    }
	    n = 0;
	}else if ( sscanf( line, "%s %s", token, label ) == 2 ) {
	    x[n] = token_indexer.key_id(token);
	    y[n] = label_indexer.key_id(label);
	    ++n;
	}else {
	    fprintf( stderr, "Invalid format: %s\n", line );
	    return -1;
	}
    }
    fclose( is );
    
    delete[] line;
    delete[] token;
    delete[] label;
    delete[] x;
    delete[] y;
    delete[] y_best;

    if ( verbose == 0 ) {
	pool.print();
	result.print();
    }
    
    return 0;
}



int load( const char* modelfile, DoubleArray& token_indexer, DoubleArray& label_indexer, HMM& hmm ) {

    FILE* is = fopen( modelfile, "rb" );
    if ( is == NULL ) {
	fprintf( stderr, "cannot open %s\n", modelfile );
	return FAILURE;
    }
    if ( token_indexer.load(is) == FAILURE ) return FAILURE;
    if ( label_indexer.load(is) == FAILURE ) return FAILURE;
    if ( hmm.load(is) == FAILURE ) return FAILURE;
    fclose( is );
    
    return SUCCESS;
}



int main( int argc, const char** argv ) {

    TestOption option;
    if ( option.read_arg( argc, argv ) == FAILURE ) return FAILURE;
    
    HMM hmm;
    DoubleArray token_indexer;
    DoubleArray label_indexer;
    Decoder<HMM> decoder;
    
    if ( load( option.modelfile(), token_indexer, label_indexer, hmm ) == FAILURE ) return FAILURE;
    decoder.init( label_indexer.key_size(), degenerate_label_num(label_indexer.key_size()), &hmm, option.decoding_option() );
    test( option.testfile(), token_indexer, label_indexer, decoder, option.verbose() );
    
    return SUCCESS;
}
