#ifndef __SD_HMM_TRAINER__
#define __SD_HMM_TRAINER__

#include <iostream>
#include <fstream>
#include <cstring> // strlen

#include "counter.h"
#include "double_array.h"
#include "hmm.h"
#include "util.h"

namespace SD {

class HMMTrainer {
  private:
    typedef HMM::Input Input;
    HMM hmm_;
    DoubleArray token_indexer_;
    DoubleArray label_indexer_;
  public:
    HMMTrainer() {};
    HMMTrainer( const HMMTrainer& obj );
    ~HMMTrainer() {};
    int preprocess( const char* trainfile );
    int train( const char* trainfile );
    int save( const char* modelfile ) const;
};


inline int HMMTrainer::preprocess( const char* trainfile ) {

    Counter token_counter;
    Counter label_counter;
    char* s = new char[BUFF_SIZE];
    char* token = new char[BUFF_SIZE];
    char* label = new char[BUFF_SIZE];
    
    FILE* is = fopen( trainfile, "r" );
    if ( is == NULL ) {
	fprintf( stderr, "cannot open %s\n", trainfile );
	return FAILURE;
    }
    setvbuf( is, NULL, _IOFBF, 1024*1024 );
    
    while ( fgets( s, BUFF_SIZE, is ) != NULL) {
	if ( s[strlen(s)-1] != '\n' ) {
	    fprintf( stderr, HERE "buffer overflow\n" );
	    return -1;
	}
	s[strlen(s)-1] = '\0';
	
	if ( strlen(s) == 0 ) {
	    continue;
	}else if ( sscanf( s, "%s %s", token, label ) == 2 ) {
	    token_counter.push( token );
	    label_counter.push( label );
	}else {
	    fprintf( stderr, "Invalid format: %s\n", s );
	    return FAILURE;
	}
    }
    fclose ( is );
    delete[] s;
    delete[] token;
    delete[] label;
    
    token_counter.prune( 2 );
    token_indexer_.build( token_counter.sorted_key() );
    label_indexer_.build( label_counter.sorted_key() );    
    
    return SUCCESS;
}


inline int HMMTrainer::train( const char* trainfile ) {

    hmm_.init( label_indexer_.key_size(), token_indexer_.key_size()+1 );

    unsigned int n = 0;
    Input* x    = new Input[MAX_LENGTH];
    Label* y    = new Label[MAX_LENGTH];
    char* token = new char[BUFF_SIZE];
    char* label = new char[BUFF_SIZE];
    char* s     = new char[BUFF_SIZE];
    
    FILE* is = fopen( trainfile, "r" );
    if ( is == NULL ) {
	fprintf( stderr, HERE "cannot open %s\n", trainfile );
	return FAILURE;
    }
    setvbuf( is, NULL, _IOFBF, 1024*1024 );
    while ( fgets( s, BUFF_SIZE, is ) != NULL) {
	if ( s[strlen(s)-1] != '\n' ) {
	    fprintf( stderr, HERE "buffer overflow\n" );
	    return FAILURE;
	}
	s[strlen(s)-1] = '\0';
	
	if ( strlen(s) == 0 ) {
	    hmm_.update( x, y, n );
	    n = 0;
	}else if ( sscanf( s, "%s %s", token, label ) == 2 ) {
	    x[n] = token_indexer_.key_id( token );
	    y[n] = label_indexer_.key_id( label );
	    n++;
	}else {
	    fprintf( stderr, HERE "invalid format: %s\n", s );
	    return FAILURE;
	}
    }
    fclose( is );
    delete[] s;
    delete[] token;
    delete[] label;
    delete[] x;
    delete[] y;
    
    hmm_.finish();
    
    unsigned int L = label_indexer_.key_size();
    unsigned int D = degenerate_label_num(L);
    unsigned int V = token_indexer_.key_size() + 1; // including an unknown word
    fprintf( stderr, "finish! (L=%d, D=%d, V=%d)\n", L, D, V );
    
    return SUCCESS;
}


inline int HMMTrainer::save( const char* filename ) const {

    FILE* os = fopen( filename, "wb" );
    if ( os == NULL ) {
	fprintf( stderr, "cannot open %s\n", filename );
	return FAILURE;
    }
    if ( token_indexer_.save(os) == FAILURE ) return FAILURE;
    if ( label_indexer_.save(os) == FAILURE ) return FAILURE;
    if ( hmm_.save(os) == FAILURE ) return FAILURE;
    fclose( os );
    return SUCCESS;
}


}

#endif

