#ifndef __HMP__
#define __HMP__

#include <iostream>
#include <fstream>
#include <vector>
#include <set>
#include <map>

#include "util.h"
#include "sparse_matrix.h"
#include "dense_matrix.h"

namespace SD {
    
class HMP {
  public:
    typedef unsigned int NodeAttribute;
    typedef unsigned int EdgeAttribute;
    struct Input {
	std::vector<NodeAttribute> node_attr;
	std::vector<EdgeAttribute> edge_attr;
    };
  private:
    unsigned int L_;   // # of original labels
    unsigned int D_;   // # of degenerate labels
    unsigned int K_;   // # of labels
    unsigned int A_;   // max. # of active labels
    unsigned int U_;   // # of node attributes
    unsigned int B_;   // # of edge attributes
    
    // Feature weights
    SparseMatrix             node_w_;  // NodeAttribute * Label
    std::vector<DenseMatrix> edge_w_;  // EdgeAttribute * Label * Label
    Real                     max_edge_w_;

    // Cached node score
    DenseMatrix              cache_;
    
    // For averaging
    SparseMatrix             node_v_;
    std::vector<DenseMatrix> edge_v_;
    unsigned int t_;
  public:
    HMP();
    HMP( const HMP& rhs );
    ~HMP() {};
    HMP& operator=( const HMP& rhs );
    
    // For training (used in hmp_train.cpp)
    void init( unsigned int L, unsigned int D, unsigned int U, unsigned int B );
    void viterbi_update( const Input* x, const Label* y_hat, const Label* y, unsigned int n );
    void staggered_update( const Input* x, const Label* y_hat, const Label* y, unsigned int n );
    void finish();
        
    // Load/save a model file
    int load( FILE* is );
    int save( FILE* os ) const;
    
    // For decoding (used in decoder.h)
    void cache_node_score( const Input* x, unsigned int n );
    Real node_score( unsigned int i, const Label& y ) const;
    Real edge_score( const Label& label1, const Label& label2, const Input& x ) const;
    Real max_edge_score() const;

    // Used in hmp_view.cpp
    Real node_attr_score( NodeAttribute attr, const Label& y ) const;
    Real edge_attr_score( EdgeAttribute attr, const Label& y1, const Label& y2 ) const;
  private:
    void estimate_ub( EdgeAttribute attr );
    void update_node_w( const Input& x, const Label& label, Real delta );
    void update_edge_w( const Input& x, const Label& label1, const Label& label2, Real delta );
};


HMP::HMP() {

    L_  = 0;
    D_  = 0;
    K_  = 0;
    A_  = 0;
    U_  = 0;
    B_  = 0;
}



void HMP::init( unsigned int L, unsigned int D, unsigned int U, unsigned int B ) {

    L_ = L;
    D_ = D;
    K_ = L + D;
    A_ = pow2(D_-1);
    U_ = U;
    B_ = B;
    t_ = 0;
    
    node_w_.init( U_ );
    edge_w_.resize( B_ );
    for ( EdgeAttribute b = 0; b < B_; ++b ) {
	edge_w_[b].init( K_, K_, ZERO );
    }
    
    node_v_.init( U_ );
    edge_v_.resize( B_ );
    for ( EdgeAttribute b = 0; b < B_; ++b ) {
	edge_v_[b].init( K_, K_, ZERO );
    }

    cache_.init( MAX_LENGTH, K_, ZERO );
}


/*
 * Update weights, if the current classification result (= y_hat) is incorrect.
 */
inline void HMP::viterbi_update( const Input* x, const Label* y_hat, const Label* y, unsigned int n ) {

    for ( unsigned int i = 0; i < n; ++i ) {
	if ( y_hat[i] != y[i] ) {
	    update_node_w( x[i], y_hat[i], -1 );
	    update_node_w( x[i], y[i], 1 );
	}
	if ( 0 < i && ( y_hat[i-1] != y[i-1] || y_hat[i] != y[i] ) ) {
	    update_edge_w( x[i-1], y_hat[i-1], y_hat[i], -1 );
	    update_edge_w( x[i-1], y[i-1], y[i], 1 );
	}
    }
    t_++;
}


inline void HMP::staggered_update( const Input* x, const Label* y_hat, const Label* y, unsigned int n ) {

    std::set<EdgeAttribute> updated_attr;
    for ( unsigned int i = 0; i < n; ++i ) {
	if ( y_hat[i] != y[i] ) {
	    update_node_w( x[i], y_hat[i], -1 );
	    update_node_w( x[i], y[i], 1 );
	}
	if ( 0 < i && ( y_hat[i-1] != y[i-1] || y_hat[i] != y[i] ) ) {
	    update_edge_w( x[i-1], y_hat[i-1], y_hat[i], -1 );
	    update_edge_w( x[i-1], y[i-1], y[i], 1 );
	    for ( std::vector<EdgeAttribute>::const_iterator it = x[i-1].edge_attr.begin(); it != x[i-1].edge_attr.end(); ++it ) {
		updated_attr.insert( *it );
	    }
	}
    }
    for ( std::set<EdgeAttribute>::const_iterator it = updated_attr.begin(); it != updated_attr.end(); ++it ) {
	estimate_ub( *it );
    }
    t_++;
}


void HMP::finish() {

    //  averaging
    node_v_ /= static_cast<Real>(t_);
    node_w_ -= node_v_;
    for ( EdgeAttribute b = 0; b < B_; ++b ) {
	edge_v_[b] /= static_cast<Real>(t_);
	edge_w_[b] -= edge_v_[b];
    }
	
    //  estimate upper bound
    for ( EdgeAttribute attr = 0; attr < B_; ++attr ) {
	estimate_ub( attr );
    }

    max_edge_w_ = ZERO;
    for ( EdgeAttribute b = 0; b < B_; ++b ) {
	max_edge_w_ += edge_w_[b].max();
    }
}


inline void HMP::estimate_ub( EdgeAttribute attr ) {

    // active label <-> degenerate label
    for ( Label y = 0; y < L_; ++y ) {
	unsigned int act_num = A_;
	edge_w_[attr][y][K_-1] = edge_w_[attr].max( y, y+1, act_num, L_ );
	edge_w_[attr][K_-1][y] = edge_w_[attr].max( act_num, L_, y, y+1 );
	for ( Label z = K_-2; L_ <= z; --z ) {
	    act_num /= 2;
	    edge_w_[attr][y][z] = std::max<Real>( edge_w_[attr][y][z+1], edge_w_[attr].max( y, y+1, act_num, 2*act_num ) );
	    edge_w_[attr][z][y] = std::max<Real>( edge_w_[attr][z+1][y], edge_w_[attr].max( act_num, 2*act_num, y, y+1 ) );
	}
    }

    // degenerate label <-> degenerate label
    for ( Label z1 = L_; z1 < K_; ++z1 ) {
	unsigned int act_num = A_;
	edge_w_[attr][z1][K_-1] = edge_w_[attr].max( z1, z1+1, act_num, L_ );
	for ( Label z2 = K_-2; L_ <= z2; --z2 ) {
	    act_num /= 2;
	    edge_w_[attr][z1][z2] = std::max<Real>( edge_w_[attr][z1][z2+1], edge_w_[attr].max( z1, z1+1, act_num, 2*act_num ) );
	}
    }
}


int HMP::load( FILE* is ) {

    if ( fread( &L_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( fread( &D_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( fread( &K_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( fread( &A_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( fread( &U_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( fread( &B_, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( fread( &max_edge_w_, sizeof( Real ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    if ( node_w_.load( is ) == FAILURE ) return FAILURE;
    
    edge_w_.resize( B_ );
    for ( EdgeAttribute b = 0; b < B_; ++b ) {
	if ( edge_w_[b].load( is ) == FAILURE ) return FAILURE;
    }
    cache_.init( MAX_LENGTH, K_, ZERO );
    
    return SUCCESS;
}


int HMP::save( FILE* os ) const {
    
    if ( fwrite( &L_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &D_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &K_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &A_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &U_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &B_, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &max_edge_w_, sizeof( Real ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( node_w_.save( os ) == FAILURE ) return FAILURE;
    for ( EdgeAttribute b = 0; b < B_; ++b ) {
	if ( edge_w_[b].save( os ) == -1 ) return FAILURE;
    }
    return SUCCESS;
}


inline void HMP::cache_node_score( const Input* x, unsigned int n ) {
    
    for ( unsigned int i = 0; i < n; ++i ) {
	DenseVector& cache = cache_[i];
	cache.fill( ZERO );
	std::vector<NodeAttribute>::const_iterator first = x[i].node_attr.begin();
	std::vector<NodeAttribute>::const_iterator last  = x[i].node_attr.end();
	for ( std::vector<NodeAttribute>::const_iterator it = first; it != last; ++it ) {
	    cache += node_w_[*it];
	}
	cache.estimate_ub( L_, D_, A_ );
    }
}


inline Real HMP::node_score( unsigned int i, const Label& y ) const {

    return cache_[i][y];
}


inline Real HMP::edge_score( const Label& y1, const Label& y2, const Input& x ) const {

    Real score = ZERO;
    std::vector<EdgeAttribute>::const_iterator first = x.edge_attr.begin();
    std::vector<EdgeAttribute>::const_iterator last  = x.edge_attr.end();
    for ( std::vector<EdgeAttribute>::const_iterator it = first; it != last; ++it ) {
	score += edge_w_[*it][y1][y2];
    }
    return score;
}


inline Real HMP::max_edge_score() const {

    return max_edge_w_;
}


inline Real HMP::node_attr_score( NodeAttribute u, const Label& y ) const {

    return node_w_[u][y];
}


inline Real HMP::edge_attr_score( EdgeAttribute e, const Label& y1, const Label& y2 ) const {

    return edge_w_[e][y1][y2];
}


inline void HMP::update_node_w( const Input& x, const Label& y, Real delta ) {

    for ( std::vector<NodeAttribute>::const_iterator it = x.node_attr.begin(); it != x.node_attr.end(); ++it ) {
	node_w_[*it][y] += delta;
	node_v_[*it][y] += static_cast<Real>(t_)*delta;
    }
}


inline void HMP::update_edge_w( const Input& x, const Label& y1, const Label& y2, Real delta ) {

    for ( std::vector<EdgeAttribute>::const_iterator it = x.edge_attr.begin(); it != x.edge_attr.end(); ++it ) {
	edge_w_[*it][y1][y2] += delta;
	edge_v_[*it][y1][y2] += static_cast<Real>(t_)*delta;
    }
}

}

#endif
