#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <math.h>
#include <cstring> // strlen

#include "hmp.h"
#include "decoder.h"
#include "double_array.h"
#include "util.h"
#include "hmp_util.h"
#include "test_option.h"
#include "result.h"
#include "timer.h"

using namespace SD;


inline void print_result( Real score, const Label* y_best, unsigned int n, const DoubleArray& label_indexer, const Decoder<HMP>& decoder ) {

    printf( "# Score=%f IterNum=%d\n", score, decoder.iter_num() );
    for ( unsigned int i = 0; i < n; ++i ) {
	printf( "%s\n", label_indexer.key(y_best[i]) );
    }
    printf( "\n" );
}


inline void print_detailed_result( Real score, unsigned int max_label_length, const Label* y_best, unsigned int n, const DoubleArray& label_indexer, const Decoder<HMP>& decoder ) {
    
    char label_format[128];
    sprintf( label_format, "%s%ds", "%", max_label_length );

    std::vector<Label> y;
    printf( "# Score=%f IterNum=%d\n", score, decoder.iter_num() );
    for ( unsigned int i = 0; i < n; ++i ) {
	decoder.debug_chart( i, y );
	for ( unsigned int j = 0; j < y.size(); ++j ) {
	    printf( " " );
	    if ( y[j] == y_best[i] ) {
		printf( "[" );
	    }else {
		printf( " " );
	    }
	    if ( y[j] < label_indexer.key_size() ) {
		printf( label_format, label_indexer.key(y[j]) );
	    }else {
		printf( label_format, "*" );
	    }
	    if ( y[j] == y_best[i] ) {
		printf( "]" );
	    }else {
		printf( " " );
	    }
	}
	printf( "\n" );
    }
    printf( "\n" );
}


int main( int argc, const char** argv ) {

    TestOption option;
    if ( option.read_arg( argc, argv ) == -1 ) return -1;

    TimerPool pool;
    Timer* timer1 = pool.push( "Total     " );
    Timer* timer2 = pool.push( "Preprocess" );
    Timer* timer3 = pool.push( "Decoding  ", "sentences" );
    
    // loading the model file
    timer1->startTimer();
    timer2->startTimer();
    
    DoubleArray attr_indexer;
    DoubleArray label_indexer;
    unsigned int node_attr_num;
    HMP         hmp;
    if ( load_model( option.modelfile(), attr_indexer, node_attr_num, label_indexer, hmp ) == -1 ) return -1;
    
    Decoder<HMP> decoder;
    unsigned int L = label_indexer.key_size();
    unsigned int D = degenerate_label_num( L );
    decoder.init( L, D, &hmp, option.decoding_option() );
    
    timer2->stopTimer();
    
    //
    FILE* is = NULL;
    if ( strcmp( option.testfile(), "-" ) == 0 ) {
	is = stdin;
    }else {
	is = fopen( option.testfile(), "r" );
    }
    if ( is == NULL ) {
	fprintf( stderr, "cannot open %s\n", option.testfile() );
	return -1;
    }
    setvbuf( is, NULL, _IOFBF, 1024*1024 );
    
    unsigned int n     = 0;
    HMP::Input* x      = new HMP::Input[MAX_LENGTH];
    Label* y      = new Label[MAX_LENGTH];
    Label* y_best = new Label[MAX_LENGTH];
    Result result;
    
    unsigned int max_label_length = 0;
    for ( Label label = 0; label < L; ++label ) {
	if ( max_label_length < strlen( label_indexer.key(label) ) ) {
	    max_label_length = strlen( label_indexer.key(label) );
	}
    }

    char* line = new char[MAX_BYTE_LENGTH];
    while ( fgets( line, MAX_BYTE_LENGTH, is ) != NULL) {
	if ( line[strlen(line)-1] != '\n' ) {
	    fprintf( stderr, HERE "buffer overflow\n" );
	    return -1;
	}
	line[strlen(line)-1] = '\0';

	
	if ( strlen(line) == 0 ) {
	    timer3->startTimer();
	    Real score = decoder.run( x, y_best, n );
	    timer3->stopTimer();
	    
	    result.update( y, y_best, n, decoder );
	    if ( option.verbose() == 1 ) {
		print_result( score, y_best, n, label_indexer, decoder );
	    }else if ( option.verbose() == 2 ) {
		print_detailed_result( score, max_label_length, y_best, n, label_indexer, decoder );
	    }
	    n = 0;
	}else if ( read_line( line, x[n], y[n], attr_indexer, node_attr_num, label_indexer ) != -1 ) {
	    ++n;
	}else {
	    fprintf( stderr, HERE "invalid format: %s\n", line );
	    return -1;
	}
    }
    fclose( is );
    timer1->stopTimer();
    
    
    if ( option.verbose() == 0 ) {
	pool.print();
	result.print();
    }

    delete[] x;
    delete[] y;
    delete[] y_best;
    delete[] line;
    
    return SUCCESS;
}
