#include <iostream>
#include <fstream>
#include <sstream>
#include <map>
#include <set>
#include <vector>
#include <string>
#include <math.h>

#include "hmp.h"
#include "double_array.h"
#include "decoder.h"
#include "hmp_util.h"
#include "util.h"
#include "counter.h"
#include "timer.h"
#include "train_option.h"

using namespace SD;

int load_example( const char* filename, std::vector<Example>& example, const DoubleArray& attr_indexer, unsigned int node_attr_num, const DoubleArray& label_indexer ) {

    FILE* is = fopen( filename, "r" );
    if ( is == NULL ) {
	fprintf( stderr, "cannot open %s\n", filename );
	return -1;
    }
    setvbuf( is, NULL, _IOFBF, 1024*1024 );

    unsigned int i = 0;
    char* line = new char[MAX_BYTE_LENGTH];
    while ( fgets( line, MAX_BYTE_LENGTH, is ) != NULL) {
	if ( line[strlen(line)-1] != '\n' ) {
	    fprintf( stderr, HERE "buffer overflow\n" );
	    return -1;
	}
	line[strlen(line)-1] = '\0';
	
	if ( strlen( line ) == 0 ) {
	    ++i;
	}else {
	    example[i].x.push_back( HMP::Input() );
	    example[i].y.push_back( Label() );
	    if ( read_line( line, example[i].x.back(), example[i].y.back(), attr_indexer, node_attr_num, label_indexer ) != -1 ) {
		continue;
	    }else {
		fprintf( stderr,"invalid format: %s\n", line );
		return -1;
	    }
	}
    }
    fclose( is );
    delete[] line;
    
    return 0;
}



void train( const std::vector<Example>& example, Algorithm algorithm, HMP& hmp, Decoder<HMP>& decoder ) {
    
    unsigned int example_num = example.size();

    std::vector<Label> y_best( MAX_LENGTH );
    
    for ( unsigned int i = 0; i < example_num; ++i ) {
	if ( i%1000 == 0 ) fprintf( stderr,"." );
	decoder.run( &example[i].x[0], &y_best[0], example[i].x.size() );
	if ( algorithm == STAGGERED ) {
	    hmp.staggered_update( &example[i].x[0], &y_best[0], &example[i].y[0], example[i].x.size() );
	}else {
	    hmp.viterbi_update( &example[i].x[0], &y_best[0], &example[i].y[0], example[i].x.size() );
	}
    }
}



int preprocess( const char* filename, DoubleArray& attr_indexer, DoubleArray& label_indexer, unsigned int& node_attr_num, unsigned int& edge_attr_num, unsigned int& example_num ) {
    
    FILE* is = fopen( filename, "r" );
    if ( is == NULL ) {
	fprintf( stderr, "cannot open %s\n", filename );
	return -1;
    }
    setvbuf( is, NULL, _IOFBF, 1024*1024 );

    example_num = 0;
    Counter label_counter;
    //Counter node_attr_set, edge_attr_set;
    std::set<std::string> node_attr_set, edge_attr_set;
    std::vector<std::string> attr_vec;
    
    char* line = new char[MAX_BYTE_LENGTH];
    while ( fgets( line, MAX_BYTE_LENGTH, is ) != NULL) {
	if ( line[strlen(line)-1] != '\n' ) {
	    fprintf( stderr, HERE "buffer overflow\n" );
	    return -1;
	}
	line[strlen(line)-1] = '\0';
	
	if ( strlen( line ) == 0 ) {
	    ++example_num;
	}else if ( read_line_preprocess( line, node_attr_set, edge_attr_set, label_counter ) == -1 ) {
	    fprintf( stderr, HERE "invalid format: %s\n", line );
	    return -1;
	}
    }
    fclose( is );
    delete[] line;

    /*
    node_attr_num = node_attr_set.size();
    edge_attr_num = edge_attr_set.size();
    attr_indexer.build( edge_attr_set.sorted_key() );
    */
    
    node_attr_num = node_attr_set.size();
    edge_attr_num = edge_attr_set.size();
    for ( std::set<std::string>::const_iterator it = node_attr_set.begin(); it != node_attr_set.end(); ++it ) {
	attr_vec.push_back( *it );
    }
    for ( std::set<std::string>::const_iterator it = edge_attr_set.begin(); it != edge_attr_set.end(); ++it ) {
	attr_vec.push_back( *it );
    }
    attr_indexer.build( attr_vec );
    
    // Sort the labels in the descending order of their frequency.
    label_indexer.build( label_counter.sorted_key() );
    
    return 0;
}


int main( int argc, const char** argv ) {

    //  Get options
    TrainOption option;
    if ( option.read_arg( argc, argv ) == FAILURE ) return FAILURE;
    
    TimerPool pool;
    Timer* timer1 = pool.push( "Total     " );
    Timer* timer2 = pool.push( "Preprosess" );
    Timer* timer3 = pool.push( "Decoding  ", "sentences" );
    
    /*
     * Preprocessing
     */
    timer1->startTimer();
    timer2->startTimer();
    
    unsigned int node_attr_num = 0;
    unsigned int edge_attr_num = 0;
    DoubleArray attr_indexer;
    DoubleArray label_indexer;
    unsigned int example_num = 0;
    if ( preprocess( option.trainfile(), attr_indexer, label_indexer, node_attr_num, edge_attr_num, example_num ) == -1 ) return -1;

    std::vector<Example> example( example_num );
    if ( load_example( option.trainfile(), example, attr_indexer, node_attr_num, label_indexer ) == -1 ) return -1;
    
    unsigned int L = label_indexer.key_size();
    unsigned int D = degenerate_label_num( L );
    timer2->stopTimer();
    
    /*
     * Training
     */
    timer3->startTimer();
    HMP          hmp;
    Decoder<HMP> decoder;
    
    hmp.init( L, D, node_attr_num, edge_attr_num );
    decoder.init( L, D, &hmp, option.decoding_option() );
    for ( unsigned int i = 0; i < option.iter_num(); ++i ) {	
	fprintf( stderr,"%d-th iteration ", i );
	train( example, option.decoding_option().algorithm, hmp, decoder );
	fprintf( stderr," done\n" );
    }
    hmp.finish();
    if ( save_model( option.modelfile(), attr_indexer, node_attr_num, label_indexer, hmp ) == -1 ) return -1;
    
    timer3->stopTimer();
    timer1->stopTimer();
    
    fprintf( stderr, "finish!\n" );
    fprintf( stderr, "(#labels=%d, #degenerate labels=%d, #node attributes=%d, #edge attributes=%d)\n\n", L, D, node_attr_num, edge_attr_num );
    
    pool.print();
    
    return SUCCESS;
}
