#ifndef __SD_HMP_TRAINER__
#define __SD_HMP_TRAINER__

#include <iostream>
#include <fstream>
#include <cstring> // strlen

#include "double_array.h"
#include "hmp.h"
#include "util.h"

namespace SD {

class HMPTrainer {
  private:
    typedef HMP::Input Input;
    struct Example {
	std::vector<Input> x;
	std::vector<Label> y;
    }; 
    DoubleArray attr_indexer_;
    DoubleArray label_indexer_;
    unsigned int node_attr_num_;
    unsigned int edge_attr_num_;
    HMP hmp_;
  public:
    HMPTrainer() {};
    HMPTrainer( const HMPTrainer& obj );
    ~HMPTrainer() {};
    int train( const char* trainfile );
    int load_example( const char* trainfile, std::vector<Example>& example );
    void one_iteration( const std::vector<Example>& example );
    int save( const char* modelfile ) const;
};


inline int HMPTrainer::preprocess( const char* filename ) {
    
    FILE* is = fopen( filename, "r" );
    if ( is == NULL ) {
	fprintf( stderr, "cannot open %s\n", filename );
	return FAILURE;
    }
    setvbuf( is, NULL, _IOFBF, 1024*1024 );

    Counter node_attr_counter, edge_attr_counter, label_counter;
    //std::set<std::string> node_attr_set, edge_attr_set;
    std::vector<std::string> attr_vec;
    
    char* line = new char[MAX_BYTE_LENGTH];
    while ( fgets( line, MAX_BYTE_LENGTH, is ) != NULL) {
	if ( line[strlen(line)-1] != '\n' ) {
	    fprintf( stderr, HERE "buffer overflow\n" );
	    return FAILURE;
	}
	line[strlen(line)-1] = '\0';
	
	if ( strlen( line ) == 0 ) {
	    continue;
	}else if ( read_line_preprocess( line, node_attr_counter, edge_attr_counter, label_counter ) == -1 ) {
	    fprintf( stderr, HERE "invalid format: %s\n", line );
	    return FAILURE;
	}
    }
    fclose( is );
    delete[] line;
    
    node_attr_num = node_attr_counter.size();
    edge_attr_num = edge_attr_counter.size();
    edge_attr_counter += node_attr_counter;
    attr_indexer.build( edge_attr_counter.sorted_key() );
    
    /*
    node_attr_num = node_attr_set.size();
    edge_attr_num = edge_attr_set.size();
    for ( std::set<std::string>::const_iterator it = node_attr_set.begin(); it != node_attr_set.end(); ++it ) {
	attr_vec.push_back( *it );
    }
    for ( std::set<std::string>::const_iterator it = edge_attr_set.begin(); it != edge_attr_set.end(); ++it ) {
	attr_vec.push_back( *it );
    }
    attr_indexer.build( attr_vec );
    */
    
    // Sort the labels in the descending order of their frequency.
    label_indexer.build( label_counter.sorted_key() );
    
    return SUCCESS;
}


inline int load_example( const char* trainfile, std::vector<Example>& example ) {

    FILE* is = fopen( trainfile, "r" );
    if ( is == NULL ) {
	fprintf( stderr, "cannot open %s\n", trainfile );
	return FAILURE;
    }
    setvbuf( is, NULL, _IOFBF, 1024*1024 );

    Example e;
    char* line = new char[MAX_BYTE_LENGTH];
    while ( fgets( line, MAX_BYTE_LENGTH, is ) != NULL) {
	if ( line[strlen(line)-1] != '\n' ) {
	    fprintf( stderr, HERE "buffer overflow\n" );
	    return FAILURE;
	}
	line[strlen(line)-1] = '\0';
	
	if ( strlen( line ) == 0 ) {
	    example.push_back( e );
	    e.clear();
	}else {
	    e.x.push_back( Input() );
	    e.y.push_back( Label() );
	    if ( read_line( line, e.x.back(), e.y.back(), attr_indexer_, node_attr_num_, label_indexer_ ) != -1 ) {
		continue;
	    }else {
		fprintf( stderr,"invalid format: %s\n", line );
		return FAILURE;
	    }
	}
    }
    fclose( is );
    delete[] line;
    
    return SUCCESS;
}


inline int train( const char* trainfile, unsigned int iter_num ) {

    std::vector<Example> example;
    load_example( trainfile, example );
    for ( unsigned int i = 0; i < iter_num; ++i ) {
	one_iteration( example );
    }
}


inline void one_iteration( const std::vector<Example>& example ) {

    std::vector<Label> y_best( MAX_LENGTH );
    for ( unsigned int i = 0; i < example_num_; ++i ) {
	if ( i%1000 == 0 ) fprintf( stderr,"." );
	decoder.run( &example[i].x[0], &y_best[0], example[i].x.size() );
	if ( algorithm == STAGGERED ) {
	    hmp.staggered_update( &example[i].x[0], &y_best[0], &example_[i].y[0], example[i].x.size() );
	}else {
	    hmp.viterbi_update( &example[i].x[0], &y_best[0], &example_[i].y[0], example[i].x.size() );
	}
    }
}


inline int save_model( const char* modelfile ) const {

    FILE* os = fopen( modelfile, "wb" );
    if ( os == NULL ) {
	fprintf( stderr, "cannot open %s\n", modelfile );
	return FAILURE;
    }
    if ( attr_indexer_.save(os) == FAILURE ) return FAILURE;
    if ( fwrite( &node_attr_num_, sizeof( unsigned int ), 1, os ) != 1 ) return FAILURE;
    if ( label_indexer_.save(os) == FAILURE ) return FAILURE;
    if ( hmp_.save(os) == FAILURE ) return FAILURE;
    fclose( os );
    return SUCCESS;
}

}

#endif

