#ifndef __SD_SPARSEVECTOR__
#define __SD_SPARSEVECTOR__

#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm> // copy, back_inserter, lower_bound

namespace SD {
    
class SparseVector {
  private:
    typedef std::pair<unsigned int,Real> Element;
    
    // Sort elements in the ascending order of the first value
    struct cmp_first {
	bool operator()( unsigned int rhs, const Element& lhs ) const;
	bool operator()( const Element& rhs, unsigned int lhs ) const;
	bool operator()( const Element& rhs, const Element& lhs ) const;
    };
    std::vector<Element> data_;
  public:
    typedef std::vector<Element>::iterator iterator;
    typedef std::vector<Element>::const_iterator const_iterator;
    SparseVector() {};
    SparseVector( const SparseVector& obj );
    ~SparseVector() {};
    SparseVector& operator=( const SparseVector& obj );
    Real operator[]( unsigned int i ) const;
    Real& operator[]( unsigned int i );
    SparseVector& operator/=( Real v );
    SparseVector& operator-=( const SparseVector& obj );
    int load( FILE* is );
    int save( FILE* os ) const;
    unsigned int size() const;
    void clear();
    iterator begin();
    iterator end();
    const_iterator begin() const;
    const_iterator end() const;
};


inline bool SparseVector::cmp_first::operator()( unsigned int rhs, const Element& lhs ) const {
    
    return rhs < lhs.first;
}


inline bool SparseVector::cmp_first::operator()( const Element& rhs, unsigned int lhs ) const {
    
    return rhs.first < lhs;
}


inline bool SparseVector::cmp_first::operator()( const Element& rhs, const Element& lhs ) const {
    
    return rhs.first < lhs.first;
}


inline SparseVector::SparseVector( const SparseVector& obj ) {

    std::copy( obj.begin(), obj.end(), std::back_inserter( data_ ) );
}


inline SparseVector& SparseVector::operator=( const SparseVector& obj ) {
    
    if ( this != &obj ) {
	data_.clear();
	std::copy( obj.begin(), obj.end(), std::back_inserter( data_ ) );
    }
    return *this;
}
    

inline Real SparseVector::operator[]( unsigned int i ) const {

    const_iterator lb = std::lower_bound( data_.begin(), data_.end(), i, cmp_first() );
    if ( lb == data_.end() || cmp_first()( i, *lb ) ) {
	return 0.0;
    }else {
	return lb->second;
    }
}


inline Real& SparseVector::operator[]( unsigned int i ) {

    iterator lb = std::lower_bound( data_.begin(), data_.end(), i, cmp_first() );
    if ( lb == data_.end() || cmp_first()( i, *lb ) ) {
	iterator it = data_.insert( lb, Element( i, 0.0 ) );
	return it->second;
    }else {
	return lb->second;
    }
}


inline SparseVector& SparseVector::operator/=( Real v ) {

    for ( iterator it = data_.begin(); it != data_.end(); ++it ) {
	it->second /= v;
    }
    return *this;
}


inline SparseVector& SparseVector::operator-=( const SparseVector& obj ) {

    if ( this == &obj ) {
	data_.clear();
    }else {
	for ( const_iterator it = obj.begin(); it != obj.end(); ++it ) {
	    (*this)[it->first] -= it->second;
	}
    }
    return *this;
}

    
int SparseVector::load( FILE* is ) {

    unsigned int size;
    if ( fread( &size, sizeof( unsigned int ), 1, is ) != 1 ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    data_.resize( size );
    if ( fread( &data_[0], sizeof( Element ), size, is ) != size ) {
	fprintf( stderr, HERE "failed to load\n" );
	return FAILURE;
    }
    return SUCCESS;
}


int SparseVector::save( FILE* os ) const {

    unsigned int size = data_.size();
    if ( fwrite( &size, sizeof( unsigned int ), 1, os ) != 1 ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    if ( fwrite( &data_[0], sizeof( Element ), size, os ) != size ) {
	fprintf( stderr, HERE "failed to save\n" );
	return FAILURE;
    }
    return SUCCESS;
}


inline unsigned int SparseVector::size() const {

    return data_.size();
}


inline void SparseVector::clear() {

    return data_.clear();
}


inline SparseVector::iterator SparseVector::begin() {

    return data_.begin();
}


inline SparseVector::iterator SparseVector::end() {

    return data_.end();
}


inline SparseVector::const_iterator SparseVector::begin() const {

    return data_.begin();
}


inline SparseVector::const_iterator SparseVector::end() const {

    return data_.end();
}


}


#endif
