#ifndef __SD_TESTOPTION__
#define __SD_TESTOPTION__

#include <cstring> // strcmp
#include <iostream>
#include <fstream>
#include <vector>

#include "util.h"

namespace SD {

class TestOption {
  private:
    const char* modelfile_;
    const char* testfile_;
    unsigned int verbose_;
    DecodingOption decoding_option_;
  public:
    TestOption();
    TestOption( const TestOption& obj );
    ~TestOption() {};
    TestOption& operator=( const TestOption& obj );
    int read_arg( unsigned int argc, const char** argv );
    const char* modelfile() const;
    const char* testfile() const;
    unsigned int beam_width() const;
    unsigned int verbose() const;
    DecodingOption decoding_option() const;
  private:
    void print_help( const char* cmd );
};


TestOption::TestOption() {

    modelfile_ = NULL;
    testfile_ = NULL;
    verbose_ = 1;
}


const char* TestOption::modelfile() const {

    return modelfile_;
}

const char* TestOption::testfile() const {

    return testfile_;
}


unsigned int TestOption::verbose() const {

    return verbose_;
}


DecodingOption TestOption::decoding_option() const {

    return decoding_option_;
}


int TestOption::read_arg( unsigned int argc, const char** argv ) {

    if ( argc < 3 ) {
	print_help( argv[0] );
	return FAILURE;
    }
    modelfile_ = argv[argc-2];
    testfile_  = argv[argc-1];
    
    for ( unsigned int i = 1; i < argc-2; ) {
	if ( strcmp( argv[i], "-b" ) == 0 && i+1 < argc-2 ) {
	    decoding_option_.beam_width = atoi( argv[i+1] );
	    i += 2;
	}else if ( strcmp( argv[i], "-v" ) == 0 && i+1 < argc-2 ) {
	    verbose_ = atoi( argv[i+1] );
	    i += 2;
	}else if ( strcmp( argv[i], "-p" ) == 0 && i+1 < argc-2 ) {
	    if ( strcmp( argv[i+1], "0" ) == 0 ) {
		decoding_option_.pruning = false;
	    }else if ( strcmp( argv[i+1], "1" ) == 0 ) {
		decoding_option_.pruning = true;
	    }else {
		print_help( argv[0] );
		return FAILURE;
	    }
	    i += 2;
	}else if ( strcmp( argv[i], "-e" ) == 0 && i+1 < argc-2 ) {
	    if ( strcmp( argv[i+1], "0" ) == 0 ) {
		decoding_option_.expansion = FLAT;
	    }else if ( strcmp( argv[i+1], "1" ) == 0 ) {
		decoding_option_.expansion = CEXP;
	    }else {
		print_help( argv[0] );
		return FAILURE;
	    }
	    i += 2;
	}else if ( strcmp( argv[i], "-l" ) == 0 && i+1 < argc-2 ) {
	    if ( strcmp( argv[i+1], "0" ) == 0 ) {
		decoding_option_.initialization = NEG_INF;
	    }else if ( strcmp( argv[i+1], "1" ) == 0 ) {
		decoding_option_.initialization = HEURISTIC;
	    }else {
		print_help( argv[0] );
		return FAILURE;
	    }
	    i += 2;
	}else if ( strcmp( argv[i], "-d" ) == 0 && i+1 < argc-2 ) {
	    if ( strcmp( argv[i+1], "0" ) == 0 ) {
		decoding_option_.algorithm = VITERBI;
	    }else if ( strcmp( argv[i+1], "1" ) == 0 ) {
		decoding_option_.algorithm = STAGGERED;
	    }else if ( strcmp( argv[i+1], "2" ) == 0 ) {
		decoding_option_.algorithm = BEAM;
	    }else if ( strcmp( argv[i+1], "3" ) == 0 ) {
		decoding_option_.algorithm = DET;
	    }else if ( strcmp( argv[i+1], "4" ) == 0 ) {
		decoding_option_.algorithm = CARPEDIEM;
	    }else {
		print_help( argv[0] );
		return FAILURE;
	    }
	    i += 2;
	}else {
	    print_help( argv[0] );
	    return FAILURE;
	}
    }
    
    if ( decoding_option_.algorithm == BEAM && decoding_option_.beam_width == 1 ) {
	decoding_option_.algorithm = DET;
    }

    return SUCCESS;
}


void TestOption::print_help( const char* cmd ) {

    fprintf( stderr, "\n" );
    fprintf( stderr, "Usage: %s [options] model test\n", cmd );
    fprintf( stderr, "\n" );
    fprintf( stderr, "Arguments:\n" );
    fprintf( stderr, "  model      -> model file\n" );
    fprintf( stderr, "  test       -> test file ('-' is the standard input)\n" );
    fprintf( stderr, "\n" );
    fprintf( stderr, "General options:\n" );
    fprintf( stderr, "  -d [0,1]   -> decoding algorithm:\n" );
    fprintf( stderr, "                0 - Viterbi decoding\n" );
    fprintf( stderr, "                1 - staggered decoding (default)\n" );
    fprintf( stderr, "  -v [0,1]   -> output format:\n" );
    fprintf( stderr, "                0  - show the computation time and accuracy\n" );
    fprintf( stderr, "                1  - show the best label sequcence and its score\n" );
    fprintf( stderr, "\n" );
    fprintf( stderr, "Options for staggered decoding:\n" );
    fprintf( stderr, "  -e [0,1]   -> expansion strategy:\n" );
    fprintf( stderr, "                0 - ordinary expansion\n" );
    fprintf( stderr, "                1 - column-wise expansion (default)\n" );
    fprintf( stderr, "  -l [0,1]   -> lower bound initialization:\n" );
    fprintf( stderr, "                0 - negative infinity\n" );
    fprintf( stderr, "                1 - score of the path obtained by left-to-right\n" );
    fprintf( stderr, "                    deteministic decoding (default)\n" );
}


}

#endif

