#ifndef __SD_TRAINOPTION__
#define __SD_TRAINOPTION__

#include <cstring>
#include <iostream>
#include <fstream>

#include "util.h"

namespace SD {

class TrainOption {
  private:
    unsigned int   iter_num_;
    const char*    modelfile_;
    const char*    trainfile_;
    DecodingOption decoding_option_;
  public:
    TrainOption();
    TrainOption( const TrainOption& obj );
    ~TrainOption() {};
    TrainOption& operator=( const TrainOption& obj );
    int read_arg( int argc, const char** argv );
    const char* modelfile() const;
    const char* trainfile() const;
    unsigned int iter_num() const;
    const DecodingOption& decoding_option() const;
  private:
    void print_help( const char* cmd );
};


TrainOption::TrainOption() {

    modelfile_ = NULL;
    trainfile_ = NULL;
}


const char* TrainOption::modelfile() const {

    return modelfile_;
}


const char* TrainOption::trainfile() const {

    return trainfile_;
}


const DecodingOption& TrainOption::decoding_option() const {

    return decoding_option_;
}


unsigned int TrainOption::iter_num() const {

    return iter_num_;
}

int TrainOption::read_arg( int argc, const char** argv ) {

    if ( argc < 3 ) {
	print_help( argv[0] );
	return FAILURE;
    }
    trainfile_ = argv[argc-2];
    modelfile_ = argv[argc-1];

    iter_num_ = 10;
    decoding_option_.expansion      = CEXP;
    decoding_option_.initialization = HEURISTIC;
    decoding_option_.algorithm      = STAGGERED;
    
    for ( int i = 1; i < argc-2; ) {
	if ( strcmp( argv[i], "-i" ) == 0 && i+1 < argc-2 ) {
	    iter_num_ = atoi( argv[i+1] );
	    i += 2;
	}else if ( strcmp( argv[i], "-e" ) == 0 && i+1 < argc-2 ) {
	    if ( strcmp( argv[i+1], "0" ) == 0 ) {
		decoding_option_.expansion = FLAT;
	    }else if ( strcmp( argv[i+1], "1" ) == 0 ) {
		decoding_option_.expansion = CEXP;
	    }else {
		print_help( argv[0] );
		return FAILURE;
	    }
	    i += 2;
	}else if ( strcmp( argv[i], "-l" ) == 0 && i+1 < argc-2 ) {
	    if ( strcmp( argv[i+1], "0" ) == 0 ) {
		decoding_option_.initialization = NEG_INF;
	    }else if ( strcmp( argv[i+1], "1" ) == 0 ) {
		decoding_option_.initialization = HEURISTIC;
	    }else {
		print_help( argv[0] );
		return FAILURE;
	    }
	    i += 2;
	}else if ( strcmp( argv[i], "-d" ) == 0 && i+1 < argc-2 ) {
	    if ( strcmp( argv[i+1], "0" ) == 0 ) {
		decoding_option_.algorithm = VITERBI;
	    }else if ( strcmp( argv[i+1], "1" ) == 0 ) {
		decoding_option_.algorithm = STAGGERED;
	    }else {
		print_help( argv[0] );
		return FAILURE;
	    }
	    i += 2;
	}else {
	    print_help( argv[0] );
	    return FAILURE;
	}
    }
    return SUCCESS;
}


void TrainOption::print_help( const char* cmd ) {

    fprintf( stderr, "\n" );
    fprintf( stderr, "Usage: %s [options] train model\n", cmd );
    fprintf( stderr, "\n" );
    fprintf( stderr, "Arguments:\n" );
    fprintf( stderr, "  train      -> training data file\n" );
    fprintf( stderr, "  model      -> model file\n" );
    fprintf( stderr, "\n" );
    fprintf( stderr, "General options:\n" );
    fprintf( stderr, "  -i [1..]   -> iteration number of perceptron (default 10)\n" );
    fprintf( stderr, "  -d [0,1]   -> decoding algorithm:\n" );
    fprintf( stderr, "                0 - Viterbi decoding\n" );
    fprintf( stderr, "                1 - staggered decoding (default)\n" );
    fprintf( stderr, "\n" );
    fprintf( stderr, "Options for staggered decoding:\n" );
    fprintf( stderr, "  -e [0,1]   -> expansion strategy:\n" );
    fprintf( stderr, "                0 - ordinary expansion\n" );
    fprintf( stderr, "                1 - column-wise expansion (default)\n" );
    fprintf( stderr, "  -l [0,1]   -> lower bound initialization:\n" );
    fprintf( stderr, "                0 - negative infinity\n" );
    fprintf( stderr, "                1 - score of the path obtained by left-to-right\n" );
    fprintf( stderr, "                    deteministic decoding (default)\n" );
}


}

#endif

